from __future__ import annotations

import functools

import pandas as pd
import pandas.testing as tm
import pytest
from numpy import testing

pytest.importorskip("psycopg2")
gpd = pytest.importorskip("geopandas")
pytest.importorskip("shapely")

pytestmark = pytest.mark.geospatial


def test_load_geodata(con):
    t = con.table("geo")
    result = t.execute()
    assert isinstance(result, gpd.GeoDataFrame)


def test_empty_select(geotable):
    expr = geotable[geotable.geo_point.geo_equals(geotable.geo_linestring)]
    result = expr.execute()
    assert len(result) == 0


def test_select_point_geodata(geotable, snapshot):
    expr = geotable["geo_point"]
    sqla_expr = expr.compile()
    compiled = str(sqla_expr)
    snapshot.assert_match(compiled, "out.sql")
    data = expr.execute()
    assert data.geom_type.iloc[0] == "Point"


def test_select_linestring_geodata(geotable, snapshot):
    expr = geotable["geo_linestring"]
    sqla_expr = expr.compile()
    compiled = str(sqla_expr)
    snapshot.assert_match(compiled, "out.sql")
    data = expr.execute()
    assert data.geom_type.iloc[0] == "LineString"


def test_select_polygon_geodata(geotable, snapshot):
    expr = geotable["geo_polygon"]
    sqla_expr = expr.compile()
    compiled = str(sqla_expr)
    snapshot.assert_match(compiled, "out.sql")
    data = expr.execute()
    assert data.geom_type.iloc[0] == "Polygon"


def test_select_multipolygon_geodata(geotable, snapshot):
    expr = geotable["geo_multipolygon"]
    sqla_expr = expr.compile()
    compiled = str(sqla_expr)
    snapshot.assert_match(compiled, "out.sql")
    data = expr.execute()
    assert data.geom_type.iloc[0] == "MultiPolygon"


def test_geo_area(geotable, gdf):
    expr = geotable.geo_multipolygon.area()
    result = expr.execute()
    expected = pd.Series([mp.area for mp in gdf.geo_multipolygon])
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_buffer(geotable, gdf):
    expr = geotable.geo_linestring.buffer(1.0)
    result = expr.execute()
    expected = pd.Series(
        [linestring.buffer(1.0).area for linestring in gdf.geo_linestring]
    )
    assert pytest.approx(result.area, abs=1e-1) == expected


def test_geo_contains(geotable):
    expr = geotable.geo_point.buffer(1.0).contains(geotable.geo_point)
    assert expr.execute().all()


def test_geo_contains_properly(geotable):
    expr = geotable.geo_point.buffer(1.0).contains_properly(geotable.geo_point)
    assert expr.execute().all()


def test_geo_covers(geotable):
    expr = geotable.geo_point.buffer(1.0).covers(geotable.geo_point)
    assert expr.execute().all()


def test_geo_covered_by(geotable):
    expr = geotable.geo_point.covered_by(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_d_fully_within(geotable):
    expr = geotable.geo_point.d_fully_within(geotable.geo_point.buffer(1.0), 2.0)
    assert expr.execute().all()


def test_geo_d_within(geotable):
    expr = geotable.geo_point.d_within(geotable.geo_point.buffer(1.0), 1.0)
    assert expr.execute().all()


def test_geo_end_point(geotable, gdf):
    expr = geotable.geo_linestring.end_point()
    result = expr.execute()
    end_point = gdf.apply(lambda x: x.geo_linestring.interpolate(1, True), axis=1)
    for a, b in zip(result, end_point):
        assert a.equals(b)


def test_geo_envelope(geotable, gdf):
    expr = geotable.geo_linestring.buffer(1.0).envelope()
    result = expr.execute()
    expected = pd.Series(
        [linestring.buffer(1.0).envelope.area for linestring in gdf.geo_linestring]
    )
    tm.assert_series_equal(result.area, expected, check_names=False)


def test_geo_within(geotable):
    expr = geotable.geo_point.within(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_disjoint(geotable):
    expr = geotable.geo_point.disjoint(geotable.geo_point)
    assert not expr.execute().any()


def test_geo_equals(geotable):
    # note: == operation just works for comparison between same datatypes
    expr = geotable.geo_point == geotable.geo_point
    assert expr.execute().all()

    expr = geotable.geo_point.geo_equals(geotable.geo_point)
    assert expr.execute().all()

    expr = geotable.geo_point.geo_equals(geotable.geo_linestring)
    assert not expr.execute().any()


def test_geo_geometry_n(geotable, gdf):
    expr = geotable.geo_multipolygon.geometry_n(1)  # PostGIS is one-indexed.
    result = expr.execute()
    first_polygon = [mp.geoms[0] for mp in gdf.geo_multipolygon]
    for a, b in zip(result, first_polygon):
        assert a.equals(b)


def test_geo_geometry_type(geotable):
    expr = geotable.geo_point.geometry_type()
    assert (expr.execute() == "ST_Point").all()
    expr = geotable.geo_multipolygon.geometry_type()
    assert (expr.execute() == "ST_MultiPolygon").all()


def test_geo_intersects(geotable):
    expr = geotable.geo_point.intersects(geotable.geo_point.buffer(1.0))
    assert expr.execute().all()


def test_geo_is_valid(geotable):
    expr = geotable.geo_point.is_valid()
    assert expr.execute().all()


def test_geo_line_locate_point(geotable):
    expr = geotable.geo_linestring.line_locate_point(geotable.geo_point)
    assert (expr.execute() == 0).all()


def test_geo_line_merge(geotable, gdf):
    expr = geotable.geo_linestring.line_merge()
    expected = gpd.GeoSeries(gdf.geo_linestring)
    tm.assert_series_equal(expr.execute().length, expected.length)


def test_geo_line_substring(geotable, gdf):
    expr = geotable.geo_linestring.line_substring(0.25, 0.75)
    result = expr.execute()
    expected = gpd.GeoSeries(gdf.geo_linestring)
    tm.assert_series_equal(expected.length / 2.0, result.length)


def test_geo_ordering_equals(geotable):
    expr = geotable.geo_point.ordering_equals(geotable.geo_point)
    assert expr.execute().all()


def test_geo_overlaps(geotable):
    expr = geotable.geo_point.overlaps(geotable.geo_point.buffer(1.0))
    assert not expr.execute().any()


def test_geo_touches(geotable):
    expr = geotable.geo_point.touches(geotable.geo_linestring)
    assert expr.execute().all()


def test_geo_distance(geotable, gdf):
    expr = geotable.geo_point.distance(geotable.geo_multipolygon.centroid())
    result = expr.execute()
    expected = pd.Series(
        [
            point.distance(mp.centroid)
            for point, mp in zip(gdf.geo_point, gdf.geo_multipolygon)
        ]
    )
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_length(geotable, gdf):
    expr = geotable.geo_linestring.length()
    result = expr.execute()
    expected = gpd.GeoSeries(gdf.geo_linestring).length
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_n_points(geotable):
    expr = geotable.geo_linestring.n_points()
    result = expr.execute()
    assert (result == 2).all()


def test_geo_perimeter(geotable):
    expr = geotable.geo_multipolygon.perimeter()
    result = expr.execute()
    # Geopandas doesn't implement perimeter, so we do a simpler check.
    assert (result > 0.0).all()


def test_geo_srid(geotable):
    expr = geotable.geo_linestring.srid()
    result = expr.execute()
    assert (result == 0).all()


def test_geo_start_point(geotable, gdf):
    expr = geotable.geo_linestring.start_point()
    result = expr.execute()
    start_point = gdf.apply(lambda x: x.geo_linestring.interpolate(0, True), axis=1)
    for a, b in zip(result, start_point):
        assert a.equals(b)


def test_geo_difference(geotable, gdf):
    expr = (
        geotable.geo_linestring.buffer(1.0)
        .difference(geotable.geo_point.buffer(0.5))
        .area()
    )
    result = expr.execute()
    expected = pd.Series(
        [
            linestring.buffer(1.0).difference(point.buffer(0.5)).area
            for linestring, point in zip(gdf.geo_linestring, gdf.geo_point)
        ]
    )
    assert pytest.approx(result, abs=1e-1) == expected


def test_geo_intersection(geotable, gdf):
    expr = (
        geotable.geo_linestring.buffer(1.0)
        .intersection(geotable.geo_point.buffer(0.5))
        .area()
    )
    result = expr.execute()
    expected = pd.Series(
        [
            linestring.buffer(1.0).intersection(point.buffer(0.5)).area
            for linestring, point in zip(gdf.geo_linestring, gdf.geo_point)
        ]
    )
    assert pytest.approx(result, abs=1e-1) == expected


def test_geo_unary_union(geotable, gdf):
    expr = geotable.geo_polygon.unary_union().area()
    expected = functools.reduce(lambda x, y: x.union(y), gdf.geo_polygon).area
    testing.assert_almost_equal(expr.execute(), expected, decimal=2)


def test_geo_union(geotable, gdf):
    expr = geotable.geo_polygon.union(geotable.geo_multipolygon).area()
    expected = pd.Series(
        [p.union(mp).area for p, mp in zip(gdf.geo_polygon, gdf.geo_multipolygon)]
    )
    tm.assert_series_equal(expr.execute(), expected, check_names=False)


def test_geo_x(geotable, gdf):
    expr = geotable.geo_point.x()
    result = expr.execute()
    expected = gpd.GeoSeries(gdf.geo_point).x
    tm.assert_series_equal(result, expected, check_names=False)


def test_geo_y(geotable, gdf):
    expr = geotable.geo_point.y()
    result = expr.execute()
    expected = gpd.GeoSeries(gdf.geo_point).y
    tm.assert_series_equal(result, expected, check_names=False)



from __future__ import annotations

import concurrent.futures
import contextlib
import itertools
import os
import subprocess
from typing import TYPE_CHECKING, Any

import oracledb
import pytest

import ibis
from ibis.backends.tests.base import ServiceBackendTest

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

ORACLE_USER = os.environ.get("IBIS_TEST_ORACLE_USER", "ibis")
ORACLE_PASS = os.environ.get("IBIS_TEST_ORACLE_PASSWORD", "ibis")
ORACLE_HOST = os.environ.get("IBIS_TEST_ORACLE_HOST", "localhost")
ORACLE_PORT = int(os.environ.get("IBIS_TEST_ORACLE_PORT", 1521))

# Creating test DB and user
# The ORACLE_DB env-var needs to be set in the compose.yaml file
# Then, after the container is running, exec in and run (from `/opt/oracle`)
# ./createAppUser user pass ORACLE_DB
# where ORACLE_DB is the same name you used in the Compose file.

# Set to ensure decimals come back as decimals
oracledb.defaults.fetch_decimals = True


class TestConf(ServiceBackendTest):
    check_dtype = False
    supports_window_operations = False
    returned_timestamp_unit = "s"
    supports_arrays = False
    supports_arrays_outside_of_select = False
    # Automatic mutate so that logical type in ibis returns as bool
    # but the backend will still do whatever it's going to do
    native_bool = False
    supports_structs = False
    supports_json = False
    rounding_method = "half_to_even"
    data_volume = "/opt/oracle/data"
    service_name = "oracle"
    deps = ("oracledb",)

    @property
    def test_files(self) -> Iterable[Path]:
        return itertools.chain(
            self.data_dir.joinpath("csv").glob("*.csv"),
            self.script_dir.joinpath("oracle").glob("*.ctl"),
        )

    def _load_data(
        self,
        *,
        user: str = ORACLE_USER,
        password: str = ORACLE_PASS,
        host: str = ORACLE_HOST,
        port: int = ORACLE_PORT,
        **_: Any,
    ) -> None:
        """Load test data into a Oracle backend instance.

        Parameters
        ----------
        data_dir
            Location of testdata
        script_dir
            Location of scripts defining schemas
        """
        database = "IBIS_TESTING"

        # suppress the exception if the user/pass/db combo already exists
        with contextlib.suppress(subprocess.CalledProcessError):
            subprocess.check_call(
                [
                    "docker",
                    "compose",
                    "exec",
                    self.service_name,
                    "./createAppUser",
                    user,
                    password,
                    database,
                ]
            )

        init_oracle_database(
            dsn=oracledb.makedsn(host, port, service_name=database),
            user=user,
            password=password,
            database=database,
            schema=self.ddl_script,
        )

        # then call sqlldr to ingest
        with concurrent.futures.ThreadPoolExecutor() as executor:
            for fut in concurrent.futures.as_completed(
                executor.submit(
                    subprocess.check_call,
                    [
                        "docker",
                        "compose",
                        "exec",
                        self.service_name,
                        "sqlldr",
                        f"{user}/{password}@{host}:{port:d}/{database}",
                        f"control=data/{ctl_file.name}",
                    ],
                    stdout=subprocess.DEVNULL,
                )
                for ctl_file in self.script_dir.joinpath("oracle").glob("*.ctl")
            ):
                fut.result()

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.oracle.connect(
            host=ORACLE_HOST,
            user=ORACLE_USER,
            password=ORACLE_PASS,
            database="IBIS_TESTING",
            port=ORACLE_PORT,
            **kw,
        )

    @staticmethod
    def format_table(name: str) -> str:
        return f'"{name}"'


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


def init_oracle_database(
    user: str,
    password: str,
    dsn: str,
    database: str,
    schema: str | None = None,
    **kwargs: Any,
) -> None:
    """Initialise `database` at `url` with `schema`.

    Parameters
    ----------
    database : str
        Name of the database to be dropped
    schema : TextIO
        File object containing schema to use
    """

    con = oracledb.connect(dsn, user=user, password=password, stmtcachesize=0)

    if schema:
        with con.cursor() as cursor:
            for stmt in schema:
                # XXX: maybe should just remove the comments in the sql file
                # so we don't end up writing an entire parser here.
                if not stmt.startswith("--"):
                    cursor.execute(stmt)



from __future__ import annotations

import pytest

import ibis


@pytest.fixture
def geo_table():
    return ibis.table({"geo1": "geometry", "geo2": "geometry"}, name="t")


def test_geospatial_unary_op_repr(geo_table):
    expr = geo_table.geo1.centroid()
    assert expr.op().name in repr(expr)


def test_geospatial_bin_op_repr(geo_table):
    expr = geo_table.geo1.d_within(geo_table.geo2, 3.0)
    assert expr.op().name in repr(expr)
    assert "distance=" in repr(expr)


def test_geospatial_bin_op_repr_no_kwarg(geo_table):
    expr = geo_table.geo1.distance(geo_table.geo2)
    assert expr.op().name in repr(expr)
    assert "distance=" not in repr(expr)
    # test that there isn't a trailing comma from empty kwargs
    assert repr(expr).endswith("r0.geo2)")



from __future__ import annotations

from ibis.backends.impala.compiler import ImpalaCompiler
from ibis.tests.expr.mocks import MockBackend


class MockImpalaConnection(MockBackend):
    compiler = ImpalaCompiler



from __future__ import annotations

import itertools
import typing
from abc import abstractmethod
from typing import Annotated, Any, Literal, Optional, TypeVar

from public import public

import ibis.expr.datashape as ds
import ibis.expr.datatypes as dt
from ibis.common.annotations import attribute
from ibis.common.collections import FrozenDict
from ibis.common.exceptions import IbisTypeError, IntegrityError, RelationError
from ibis.common.grounds import Concrete
from ibis.common.patterns import Between, InstanceOf
from ibis.common.typing import Coercible, VarTuple
from ibis.expr.operations.core import Alias, Column, Node, Scalar, Value
from ibis.expr.operations.sortkeys import SortKey
from ibis.expr.schema import Schema
from ibis.formats import TableProxy  # noqa: TCH001

T = TypeVar("T")

Unaliased = Annotated[T, ~InstanceOf(Alias)]
NonSortKey = Annotated[T, ~InstanceOf(SortKey)]


@public
class Relation(Node, Coercible):
    @classmethod
    def __coerce__(cls, value):
        from ibis.expr.types import Table

        if isinstance(value, Relation):
            return value
        elif isinstance(value, Table):
            return value.op()
        else:
            raise TypeError(f"Cannot coerce {value!r} to a Relation")

    @property
    @abstractmethod
    def values(self) -> FrozenDict[str, Value]:
        """A mapping of column names to expressions which build up the relation.

        This attribute is heavily used in rewrites as well as during field
        dereferencing in the API layer. The returned expressions must only
        originate from parent relations, depending on the relation type.
        """

    @property
    @abstractmethod
    def schema(self) -> Schema:
        """The schema of the relation.

        All relations must have a well-defined schema.
        """
        ...

    @property
    def fields(self) -> FrozenDict[str, Column]:
        """A mapping of column names to fields of the relation.

        This calculated property shouldn't be overridden in subclasses since it
        is mostly used for convenience.
        """
        return FrozenDict({k: Field(self, k) for k in self.schema})

    def to_expr(self):
        from ibis.expr.types import Table

        return Table(self)


@public
class Field(Value):
    rel: Relation
    name: str

    shape = ds.columnar

    def __init__(self, rel, name):
        if name not in rel.schema:
            columns_formatted = ", ".join(map(repr, rel.schema.names))
            raise IbisTypeError(
                f"Column {name!r} is not found in table. "
                f"Existing columns: {columns_formatted}."
            )
        super().__init__(rel=rel, name=name)

    @attribute
    def dtype(self):
        return self.rel.schema[self.name]

    @attribute
    def relations(self):
        return frozenset({self.rel})


def _check_integrity(values, allowed_parents):
    for value in values:
        for rel in value.relations:
            if rel not in allowed_parents:
                raise IntegrityError(
                    f"Cannot add {value!r} to projection, they belong to another relation"
                )


@public
class Project(Relation):
    parent: Relation
    values: FrozenDict[str, NonSortKey[Unaliased[Value]]]

    def __init__(self, parent, values):
        _check_integrity(values.values(), {parent})
        super().__init__(parent=parent, values=values)

    @attribute
    def schema(self):
        return Schema({k: v.dtype for k, v in self.values.items()})


class Simple(Relation):
    parent: Relation

    @attribute
    def values(self):
        return self.parent.fields

    @attribute
    def schema(self):
        return self.parent.schema


# TODO(kszucs): remove in favor of View
@public
class SelfReference(Simple):
    _uid_counter = itertools.count()

    identifier: Optional[int] = None

    def __init__(self, parent, identifier):
        if identifier is None:
            identifier = next(self._uid_counter)
        super().__init__(parent=parent, identifier=identifier)

    @attribute
    def values(self):
        return FrozenDict()


JoinKind = Literal[
    "inner",
    "left",
    "right",
    "outer",
    "asof",
    "semi",
    "anti",
    "any_inner",
    "any_left",
    "cross",
]


@public
class JoinTable(Simple):
    index: int


@public
class JoinLink(Node):
    how: JoinKind
    table: JoinTable
    predicates: VarTuple[Value[dt.Boolean]]


@public
class JoinChain(Relation):
    first: JoinTable
    rest: VarTuple[JoinLink]
    values: FrozenDict[str, Unaliased[Value]]

    def __init__(self, first, rest, values):
        allowed_parents = {first}
        assert first.index == 0
        for join in rest:
            assert join.table.index == len(allowed_parents)
            allowed_parents.add(join.table)
            _check_integrity(join.predicates, allowed_parents)
        _check_integrity(values.values(), allowed_parents)
        super().__init__(first=first, rest=rest, values=values)

    @property
    def length(self):
        return len(self.rest) + 1

    @attribute
    def schema(self):
        return Schema({k: v.dtype.copy(nullable=True) for k, v in self.values.items()})

    def to_expr(self):
        import ibis.expr.types as ir

        return ir.Join(self)


@public
class Sort(Simple):
    keys: VarTuple[SortKey]

    def __init__(self, parent, keys):
        _check_integrity(keys, {parent})
        super().__init__(parent=parent, keys=keys)


@public
class Filter(Simple):
    predicates: VarTuple[Value[dt.Boolean]]

    def __init__(self, parent, predicates):
        from ibis.expr.rewrites import ReductionLike

        for pred in predicates:
            if pred.find(ReductionLike, filter=Value):
                raise IntegrityError(
                    f"Cannot add {pred!r} to filter, it is a reduction which "
                    "must be converted to a scalar subquery first"
                )
            if pred.relations and parent not in pred.relations:
                raise IntegrityError(
                    f"Cannot add {pred!r} to filter, they belong to another relation"
                )
        super().__init__(parent=parent, predicates=predicates)


@public
class Limit(Simple):
    # TODO(kszucs): dynamic limit should contain ScalarSubqueries rather than
    # plain scalar values
    n: typing.Union[int, Scalar[dt.Integer], None] = None
    offset: typing.Union[int, Scalar[dt.Integer]] = 0


@public
class Aggregate(Relation):
    parent: Relation
    groups: FrozenDict[str, Unaliased[Column]]
    metrics: FrozenDict[str, Unaliased[Scalar]]

    def __init__(self, parent, groups, metrics):
        _check_integrity(groups.values(), {parent})
        _check_integrity(metrics.values(), {parent})
        if duplicates := groups.keys() & metrics.keys():
            raise RelationError(
                f"Cannot add {duplicates} to aggregate, they are already in the groupby"
            )
        super().__init__(parent=parent, groups=groups, metrics=metrics)

    @attribute
    def values(self):
        return FrozenDict({**self.groups, **self.metrics})

    @attribute
    def schema(self):
        return Schema({k: v.dtype for k, v in self.values.items()})


@public
class Set(Relation):
    left: Relation
    right: Relation
    distinct: bool = False
    values = FrozenDict()

    def __init__(self, left, right, **kwargs):
        # convert to dictionary first, to get key-unordered comparison semantics
        if dict(left.schema) != dict(right.schema):
            raise RelationError("Table schemas must be equal for set operations")
        elif left.schema.names != right.schema.names:
            # rewrite so that both sides have the columns in the same order making it
            # easier for the backends to implement set operations
            cols = {name: Field(right, name) for name in left.schema.names}
            right = Project(right, cols)
        super().__init__(left=left, right=right, **kwargs)

    @attribute
    def schema(self):
        return self.left.schema


@public
class Union(Set):
    pass


@public
class Intersection(Set):
    pass


@public
class Difference(Set):
    pass


@public
class PhysicalTable(Relation):
    name: str
    values = FrozenDict()


@public
class Namespace(Concrete):
    database: Optional[str] = None
    schema: Optional[str] = None


@public
class UnboundTable(PhysicalTable):
    schema: Schema
    namespace: Namespace = Namespace()


@public
class DatabaseTable(PhysicalTable):
    schema: Schema
    source: Any
    namespace: Namespace = Namespace()


@public
class InMemoryTable(PhysicalTable):
    schema: Schema
    data: TableProxy


@public
class SQLQueryResult(Relation):
    """A table sourced from the result set of a select query."""

    query: str
    schema: Schema
    source: Any
    values = FrozenDict()


@public
class View(PhysicalTable):
    """A view created from an expression."""

    # TODO(kszucs): rename it to parent
    child: Relation

    @attribute
    def schema(self):
        return self.child.schema


@public
class SQLStringView(Relation):
    """A view created from a SQL string."""

    child: Relation
    query: str
    schema: Schema
    values = FrozenDict()


@public
class DummyTable(Relation):
    values: FrozenDict[str, Value]

    @attribute
    def schema(self):
        return Schema({k: v.dtype for k, v in self.values.items()})


@public
class FillNa(Simple):
    """Fill null values in the table."""

    replacements: typing.Union[Value[dt.Numeric | dt.String], FrozenDict[str, Any]]


@public
class DropNa(Simple):
    """Drop null values in the table."""

    how: typing.Literal["any", "all"]
    subset: Optional[VarTuple[Column]] = None


@public
class Sample(Simple):
    """Sample performs random sampling of records in a table."""

    fraction: Annotated[float, Between(0, 1)]
    method: typing.Literal["row", "block"]
    seed: typing.Union[int, None] = None


@public
class Distinct(Simple):
    """Distinct is a table-level unique-ing operation."""


# TODO(kszucs): support t.select(*t) syntax by implementing Table.__iter__()



import ibis


alltypes = ibis.table(
    name="alltypes",
    schema={
        "float_col": "float32",
        "timestamp_col": "timestamp",
        "int_col": "int32",
        "string_col": "string",
    },
)
param = ibis.param("timestamp")
f = alltypes.filter(alltypes.timestamp_col < param.name("my_param"))
agg = f.aggregate([f.float_col.sum().name("foo")], by=[f.string_col])

result = agg.foo.count()



from __future__ import annotations

import contextlib
import uuid

import pytest

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt

RAW_TEST_UUID = "08f48812-7948-4718-96c7-27fa6a398db6"
TEST_UUID = uuid.UUID(RAW_TEST_UUID)

UUID_BACKEND_TYPE = {
    "bigquery": "STRING",
    "clickhouse": "Nullable(UUID)",
    "duckdb": "UUID",
    "exasol": "UUID",
    "flink": "CHAR(36) NOT NULL",
    "impala": "STRING",
    "postgres": "uuid",
    "risingwave": "character varying",
    "snowflake": "VARCHAR",
    "sqlite": "text",
    "trino": "uuid",
}


@pytest.mark.notimpl(["polars"], raises=NotImplementedError)
def test_uuid_literal(con, backend):
    backend_name = backend.name()

    expr = ibis.literal(RAW_TEST_UUID, type=dt.uuid)
    result = con.execute(expr)

    assert result == TEST_UUID

    with contextlib.suppress(com.OperationNotDefinedError):
        assert con.execute(expr.typeof()) == UUID_BACKEND_TYPE[backend_name]



from __future__ import annotations

import datetime
import math
from abc import ABC, abstractmethod
from collections import defaultdict

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.flink.datatypes import FlinkType
from ibis.common.temporal import IntervalUnit, normalize_timezone
from ibis.util import convert_unit

# For details on what precisions Flink SQL interval types support, see
# https://nightlies.apache.org/flink/flink-docs-release-1.17/docs/dev/table/types/#interval-year-to-month
# https://nightlies.apache.org/flink/flink-docs-release-1.17/docs/dev/table/types/#interval-day-to-second
MIN_ALLOWED_PRECISION = {
    IntervalUnit.YEAR: 1,
    IntervalUnit.MONTH: 2,
    IntervalUnit.DAY: 1,
    IntervalUnit.HOUR: 2,
    IntervalUnit.MINUTE: 2,
    IntervalUnit.SECOND: 0,
}

MAX_ALLOWED_PRECISION = {
    IntervalUnit.YEAR: 4,
    IntervalUnit.MONTH: 2,
    IntervalUnit.DAY: 6,
    IntervalUnit.HOUR: 2,
    IntervalUnit.MINUTE: 2,
    IntervalUnit.SECOND: 9,
}

MICROSECONDS_IN_UNIT = {
    unit: datetime.timedelta(**{unit.plural: 1}).total_seconds() * 10**6
    for unit in [
        IntervalUnit.DAY,
        IntervalUnit.HOUR,
        IntervalUnit.MINUTE,
        IntervalUnit.SECOND,
    ]
}
MICROSECONDS_IN_UNIT[IntervalUnit.MONTH] = (
    30 * datetime.timedelta(days=1).total_seconds() * 10**6
)
MICROSECONDS_IN_UNIT[IntervalUnit.YEAR] = (
    365 * datetime.timedelta(days=1).total_seconds() * 10**6
)


def _calculate_precision(interval_value: int) -> int:
    """Calculate interval precision.

    FlinkSQL interval data types use leading precision and fractional-
    seconds precision. Because the leading precision defaults to 2, we need to
    specify a different precision when the value exceeds 2 digits.

    (see
    https://learn.microsoft.com/en-us/sql/odbc/reference/appendixes/interval-literals)
    """
    # log10(interval_value) + 1 is equivalent to len(str(interval_value)), but is significantly
    # faster and more memory-efficient
    if interval_value == 0:
        return 0
    if interval_value < 0:
        raise ValueError(
            f"Expecting value to be a non-negative integer, got {interval_value}"
        )
    return int(math.log10(interval_value)) + 1


def _format_value_with_precision(value: int, precision: int) -> str:
    """Format value so that it fills a specified precision."""
    return str(value).zfill(precision)


def format_precision(precision: int, unit: IntervalUnit) -> str:
    """Format precision values in Flink SQL."""
    if precision > MAX_ALLOWED_PRECISION[unit]:
        raise ValueError(
            f"{precision} is bigger than the allowed precision for {unit} ({MAX_ALLOWED_PRECISION[unit]})"
        )

    return "" if precision <= 2 else f"({precision})"


class FlinkIntervalType(ABC):
    """Abstract Base Class for Flink interval type.

    Flink supports only two types of temporal intervals: day-time intervals with up to nanosecond
    granularity or year-month intervals with up to month granularity.

    This Abstract Base Class provides functionality so that a given IntervalType instance can be
    translated appropriately into Flink SQL.
    """

    def __init__(self, value: int, unit: str) -> None:
        self.value = value
        self.unit = unit
        self.interval_segments = self._convert_to_combined_units()
        self.precisions = self._calculate_precisions()

    @classmethod
    @property
    @abstractmethod
    def units(self):
        ...

    @classmethod
    @property
    @abstractmethod
    def factors(self):
        ...

    @abstractmethod
    def _convert_to_highest_resolution(self):
        ...

    def _convert_to_combined_units(self) -> dict:
        converted_total = self._convert_to_highest_resolution()
        interval_segments = defaultdict(int)

        rem = converted_total
        for unit, factor in zip(self.units[:-1], self.factors):
            q, rem = divmod(rem, factor)
            interval_segments[unit] = int(q)
            if rem == 0:
                break
        if rem > 0:
            interval_segments[self.units[-1]] = rem
        return interval_segments

    @abstractmethod
    def _calculate_precisions(self) -> dict:
        ...

    @abstractmethod
    def format_as_string(self, interval_segments: dict, precisions: dict) -> str:
        ...


class YearsToMonthsInterval(FlinkIntervalType):
    units = (IntervalUnit.YEAR, IntervalUnit.MONTH)
    factors = (12,)

    def _convert_to_highest_resolution(self) -> int:
        return convert_unit(self.value, self.unit, to=IntervalUnit.MONTH.value)

    def _calculate_precisions(self) -> dict:
        precisions = {}
        for unit in self.units:
            value = self.interval_segments[unit]
            prec = _calculate_precision(value)
            precisions[unit] = max(prec, 2)
        return precisions

    def format_as_string(self) -> str:
        years = self.interval_segments[IntervalUnit.YEAR]
        months = self.interval_segments[IntervalUnit.MONTH]
        return (
            f"'{_format_value_with_precision(years, self.precisions[IntervalUnit.YEAR])}"
            f"-{_format_value_with_precision(months, self.precisions[IntervalUnit.MONTH])}' YEAR"
            f"{format_precision(self.precisions[IntervalUnit.YEAR], IntervalUnit.YEAR)} "
            "TO MONTH"
            f"{format_precision(self.precisions[IntervalUnit.MONTH], IntervalUnit.MONTH)}"
        )


class DaysToSecondsInterval(FlinkIntervalType):
    units = (
        IntervalUnit.DAY,
        IntervalUnit.HOUR,
        IntervalUnit.MINUTE,
        IntervalUnit.SECOND,
        IntervalUnit.MICROSECOND,
    )
    factors = (86400 * 10**6, 3600 * 10**6, 60 * 10**6, 10**6)

    def _convert_to_highest_resolution(self) -> int:
        return convert_unit(self.value, self.unit, to=IntervalUnit.MICROSECOND.value)

    def _calculate_precisions(self) -> dict:
        precisions = {}
        for unit in self.units:
            value = self.interval_segments[unit]
            if unit != IntervalUnit.MICROSECOND:
                prec = _calculate_precision(value)
                precisions[unit] = max(prec, 2)
            else:
                precisions[IntervalUnit.MICROSECOND] = max(prec, 6)
        return precisions

    def format_as_string(self) -> str:
        days = self.interval_segments[IntervalUnit.DAY]
        hours = self.interval_segments[IntervalUnit.HOUR]
        minutes = self.interval_segments[IntervalUnit.MINUTE]
        seconds = self.interval_segments[IntervalUnit.SECOND]
        microseconds = self.interval_segments[IntervalUnit.MICROSECOND]

        return (
            f"'{_format_value_with_precision(days, self.precisions[IntervalUnit.DAY])} "
            f"{_format_value_with_precision(hours, self.precisions[IntervalUnit.HOUR])}:"
            f"{_format_value_with_precision(minutes, self.precisions[IntervalUnit.MINUTE])}:"
            f"{_format_value_with_precision(seconds, self.precisions[IntervalUnit.SECOND])}."
            f"{_format_value_with_precision(microseconds, self.precisions[IntervalUnit.MICROSECOND])}' "
            f"DAY{format_precision(self.precisions[IntervalUnit.DAY], IntervalUnit.DAY)} "
            "TO SECOND"
            f"{format_precision(self.precisions[IntervalUnit.SECOND], IntervalUnit.SECOND)}"
        )


def _translate_interval(value, dtype):
    """Convert interval to Flink SQL type.

    Flink supports only two types of temporal intervals: day-time intervals with up to nanosecond
    granularity or year-month intervals with up to month granularity.

    An interval of year-month consists of +years-months with values ranging from -9999-11 to +9999-11.
    An interval of day-time consists of +days hours:minutes:seconds.fractional with values ranging from
    -999999 23:59:59.999999999 to +999999 23:59:59.999999999.

    The value representation is the same for all types of resolutions.

    For example, an interval of months of 50 is always represented in an interval-of-years-to-months
    format (with default year precision): +04-02; an interval of seconds of 70 is always represented in
    an interval-of-days-to-seconds format (with default precisions): +00 00:01:10.000000.
    """
    if dtype.unit in YearsToMonthsInterval.units:
        interval = YearsToMonthsInterval(value, dtype.unit.value)
    else:
        interval = DaysToSecondsInterval(value, dtype.unit.value)

    interval_segments = interval.interval_segments
    nonzero_interval_segments = {k: v for k, v in interval_segments.items() if v != 0}

    # YEAR, MONTH, DAY, HOUR, MINUTE, SECOND
    if len(nonzero_interval_segments) == 1:
        unit = next(iter(nonzero_interval_segments))
        value = nonzero_interval_segments[unit]
        precision = _calculate_precision(value)
        return f"'{value}' {unit.name}{format_precision(precision, unit)}"

    # YEAR TO MONTH, DAY TO SECOND
    return interval.format_as_string()


def translate_literal(op: ops.Literal) -> str:
    value = op.value
    dtype = op.dtype

    if value is None:
        if dtype.is_null():
            return "NULL"
        return f"CAST(NULL AS {FlinkType.from_ibis(dtype)!s})"

    if dtype.is_boolean():
        # TODO(chloeh13q): Flink supports a third boolean called "UNKNOWN"
        return "TRUE" if value else "FALSE"
    elif dtype.is_string():
        quoted = value.replace("'", "''")
        return f"'{quoted}'"
    elif dtype.is_binary():
        return f"x'{value.hex()}'"
    elif dtype.is_date():
        if isinstance(value, datetime.date):
            value = value.isoformat()
        return repr(value)
    elif dtype.is_numeric():
        if math.isnan(value):
            raise ValueError("NaN is not supported in Flink SQL")
        elif math.isinf(value):
            raise ValueError("Infinity is not supported in Flink SQL")
        elif dtype.is_decimal():
            # When PyFlink infers schema from `decimal.Decimal` objects,
            # it will be `DecimalType(38, 18)`.
            # https://github.com/apache/flink/blob/release-1.17.1/flink-python/pyflink/table/types.py#L336-L337
            precision = 38 if dtype.precision is None else dtype.precision
            scale = 18 if dtype.scale is None else dtype.scale

            if precision > 38:
                raise ValueError("The precision can be up to 38 in Flink")

            return f"CAST({value} AS DECIMAL({precision}, {scale}))"
        return f"CAST({value} AS {FlinkType.from_ibis(dtype)!s})"
    elif dtype.is_timestamp():
        # TODO(chloeh13q): support timestamp with local timezone
        assert isinstance(value, datetime.datetime)
        # datetime.datetime only supports resolution up to microseconds, even
        # though Flink supports fractional precision up to 9 digits. We will
        # need to use numpy or pandas datetime types for higher resolutions.
        #
        if dtype.timezone is not None:
            value = value.astimezone(normalize_timezone("UTC"))

        # remove timezone information without altering the ISO output
        # except for removing the UTC offset
        #
        # format to ISO 8601 without the T character
        value = value.replace(tzinfo=None).isoformat(sep=" ")

        if (tz := dtype.timezone) is not None:
            return f"TO_TIMESTAMP(CONVERT_TZ({value!r}, 'UTC', {tz!r}))"
        else:
            return f"TIMESTAMP {value!r}"
    elif dtype.is_time():
        return f"TIME '{value}'"
    elif dtype.is_interval():
        return f"INTERVAL {_translate_interval(value, dtype)}"
    elif dtype.is_uuid():
        return translate_literal(ops.Literal(str(value), dtype=dt.str))
    elif dtype.is_array():
        return f"ARRAY{list(value)}"

    raise NotImplementedError(f"No translation rule for {dtype}")



from __future__ import annotations

import pytest
from pyflink.util.exceptions import TableException
from pytest import param

import ibis
from ibis.backends.tests.errors import Py4JJavaError


@pytest.mark.xfail(raises=TableException)
def test_window_requires_order_by(con):
    t = con.tables.functional_alltypes
    expr = t.mutate(t.double_col - t.double_col.mean())
    con.execute(expr)


@pytest.mark.xfail(raises=TableException)
def test_window_does_not_support_multiple_order_by(con):
    t = con.tables.functional_alltypes
    expr = t.double_col.sum().over(rows=(-1, 1), order_by=[t.timestamp_col, t.int_col])
    con.execute(expr)


@pytest.mark.parametrize(
    "window",
    [
        param(
            {"rows": (-1, 1)},
            id="bounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"rows": (-1, None)},
            id="unbounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"rows": (-500, 1)},
            id="casted_bounded_rows_following",
            marks=[pytest.mark.xfail(raises=TableException)],
        ),
        param(
            {"range": (-1000, 0)},
            id="int_range",
            marks=[pytest.mark.xfail(raises=Py4JJavaError)],
        ),
    ],
)
def test_window_invalid_start_end(con, window):
    t = con.tables.functional_alltypes
    expr = t.int_col.sum().over(**window, order_by=t.timestamp_col)
    con.execute(expr)


def test_range_window(con, snapshot, simple_table):
    expr = simple_table.f.sum().over(
        range=(-ibis.interval(minutes=500), 0), order_by=simple_table.f
    )
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


def test_rows_window(con, snapshot, simple_table):
    expr = simple_table.f.sum().over(rows=(-1000, 0), order_by=simple_table.f)
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")



from __future__ import annotations

from collections.abc import Mapping

import numpy as np
import pandas as pd
import pandas.testing as tm
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt

pytestmark = [
    pytest.mark.never(["mysql", "sqlite", "mssql"], reason="No struct support"),
    pytest.mark.notyet(["impala"]),
    pytest.mark.notimpl(["datafusion", "druid", "oracle", "exasol"]),
]


@pytest.mark.notimpl(["dask"])
@pytest.mark.parametrize(
    ("field", "expected"),
    [
        param(
            "a",
            [1.0, 2.0, 2.0, 3.0, 3.0, np.nan, np.nan],
            id="a",
            marks=pytest.mark.notimpl(["snowflake"]),
        ),
        param(
            "b", ["apple", "banana", "banana", "orange", "orange", None, None], id="b"
        ),
        param(
            "c",
            [2, 2, 3, 3, 4, np.nan, np.nan],
            id="c",
            marks=pytest.mark.notimpl(["snowflake"]),
        ),
    ],
)
def test_single_field(struct, field, expected):
    expr = struct.select(field=lambda t: t.abc[field]).order_by("field")
    result = expr.execute()
    tm.assert_series_equal(result.field, pd.Series(expected, name="field"))


@pytest.mark.notimpl(["dask"])
def test_all_fields(struct, struct_df):
    result = struct.abc.execute()
    expected = struct_df.abc

    assert {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in result
    } == {
        row if not isinstance(row, Mapping) else tuple(row.items()) for row in expected
    }


_SIMPLE_DICT = dict(a=1, b="2", c=3.0)
_STRUCT_LITERAL = ibis.struct(
    _SIMPLE_DICT,
    type="struct<a: int64, b: string, c: float64>",
)
_NULL_STRUCT_LITERAL = ibis.NA.cast("struct<a: int64, b: string, c: float64>")


@pytest.mark.notimpl(["postgres", "risingwave"])
@pytest.mark.parametrize("field", ["a", "b", "c"])
@pytest.mark.notyet(
    ["flink"], reason="flink doesn't support creating struct columns from literals"
)
def test_literal(backend, con, field):
    query = _STRUCT_LITERAL[field]
    dtype = query.type().to_pandas()
    result = pd.Series([con.execute(query)], dtype=dtype)
    result = result.replace({np.nan: None})
    expected = pd.Series([_SIMPLE_DICT[field]])
    backend.assert_series_equal(result, expected.astype(dtype))


@pytest.mark.notimpl(["postgres"])
@pytest.mark.parametrize("field", ["a", "b", "c"])
@pytest.mark.notyet(
    ["clickhouse"], reason="clickhouse doesn't support nullable nested types"
)
@pytest.mark.notyet(
    ["flink"], reason="flink doesn't support creating struct columns from literals"
)
def test_null_literal(backend, con, field):
    query = _NULL_STRUCT_LITERAL[field]
    result = pd.Series([con.execute(query)])
    result = result.replace({np.nan: None})
    expected = pd.Series([None], dtype="object")
    backend.assert_series_equal(result, expected)


@pytest.mark.notimpl(["dask", "pandas", "postgres", "risingwave"])
@pytest.mark.notyet(
    ["flink"], reason="flink doesn't support creating struct columns from literals"
)
def test_struct_column(alltypes, df):
    t = alltypes
    expr = t.select(s=ibis.struct(dict(a=t.string_col, b=1, c=t.bigint_col)))
    assert expr.s.type() == dt.Struct(dict(a=dt.string, b=dt.int8, c=dt.int64))
    result = expr.execute()
    expected = pd.DataFrame(
        {"s": [dict(a=a, b=1, c=c) for a, c in zip(df.string_col, df.bigint_col)]}
    )
    tm.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["dask", "pandas", "postgres", "risingwave", "polars"])
@pytest.mark.notyet(
    ["flink"], reason="flink doesn't support creating struct columns from collect"
)
def test_collect_into_struct(alltypes):
    from ibis import _

    t = alltypes
    expr = (
        t[_.string_col.isin(("0", "1"))]
        .group_by(group="string_col")
        .agg(
            val=lambda t: ibis.struct(
                dict(key=t.bigint_col.collect().cast("!array<int64>"))
            )
        )
    )
    result = expr.execute()
    assert result.shape == (2, 2)
    assert set(result.group) == {"0", "1"}
    val = result.val
    assert len(val.loc[result.group == "0"].iat[0]["key"]) == 730
    assert len(val.loc[result.group == "1"].iat[0]["key"]) == 730



"""Various traversal utilities for the expression graph."""
from __future__ import annotations

from abc import abstractmethod
from collections import deque
from collections.abc import Iterable, Iterator, KeysView, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union

from ibis.common.bases import Hashable
from ibis.common.collections import frozendict
from ibis.common.patterns import NoMatch, Pattern
from ibis.common.typing import _ClassInfo
from ibis.util import experimental

if TYPE_CHECKING:
    from typing_extensions import Self

    N = TypeVar("N")


Finder = Callable[["Node"], bool]
FinderLike = Union[Finder, Pattern, _ClassInfo]

Replacer = Callable[["Node", dict["Node", Any]], "Node"]
ReplacerLike = Union[Replacer, Pattern, Mapping]


def _flatten_collections(node: Any) -> Iterator[N]:
    """Flatten collections of nodes into a single iterator.

    We treat common collection types inherently traversable (e.g. list, tuple, dict)
    but as undesired in a graph representation, so we traverse them implicitly.

    Parameters
    ----------
    node
        Flattaneble object.

    Returns
    -------
    A flat generator of the filtered nodes.

    Examples
    --------
    >>> from ibis.common.grounds import Concrete
    >>> from ibis.common.graph import Node
    >>>
    >>> class MyNode(Concrete, Node):
    ...     number: int
    ...     string: str
    ...     children: tuple[Node, ...]
    >>> a = MyNode(4, "a", ())
    >>>
    >>> b = MyNode(3, "b", ())
    >>> c = MyNode(2, "c", (a, b))
    >>> d = MyNode(1, "d", (c,))
    >>>
    >>> assert list(_flatten_collections(a)) == [a]
    >>> assert list(_flatten_collections((c,))) == [c]
    >>> assert list(_flatten_collections([a, b, (c, a)])) == [a, b, c, a]

    """
    if isinstance(node, Node):
        yield node
    elif isinstance(node, (tuple, list)):
        for item in node:
            yield from _flatten_collections(item)
    elif isinstance(node, (dict, frozendict)):
        for value in node.values():
            yield from _flatten_collections(value)


def _recursive_lookup(obj: Any, dct: dict) -> Any:
    """Recursively replace objects in a nested structure with values from a dict.

    Since we treat common collection types inherently traversable, so we need to
    traverse them implicitly and replace the values given a result mapping.

    Parameters
    ----------
    obj
        Object to replace.
    dct
        Mapping of objects to replace with their values.

    Returns
    -------
    Object with replaced values.

    Examples
    --------
    >>> from ibis.common.grounds import Concrete
    >>> from ibis.common.graph import Node
    >>>
    >>> class MyNode(Concrete, Node):
    ...     number: int
    ...     string: str
    ...     children: tuple[Node, ...]
    >>> a = MyNode(4, "a", ())
    >>>
    >>> b = MyNode(3, "b", ())
    >>> c = MyNode(2, "c", (a, b))
    >>> d = MyNode(1, "d", (c,))
    >>>
    >>> dct = {a: "A", b: "B"}
    >>> _recursive_lookup(a, dct)
    'A'
    >>> _recursive_lookup((a, b), dct)
    ('A', 'B')
    >>> _recursive_lookup({1: a, 2: b}, dct)
    {1: 'A', 2: 'B'}
    >>> _recursive_lookup((a, frozendict({1: c})), dct)
    ('A', {1: MyNode(number=2, ...)})

    """
    if isinstance(obj, Node):
        return dct.get(obj, obj)
    elif isinstance(obj, (tuple, list)):
        return tuple(_recursive_lookup(o, dct) for o in obj)
    elif isinstance(obj, (dict, frozendict)):
        return {k: _recursive_lookup(v, dct) for k, v in obj.items()}
    else:
        return obj


def _coerce_finder(obj: FinderLike, context: Optional[dict] = None) -> Finder:
    """Coerce an object into a callable finder function.

    Parameters
    ----------
    obj
        A callable accepting the node, a pattern or a type to match on.
    context
        Optional context to use if the finder is a pattern.

    Returns
    -------
    A callable finder function which can be used to match nodes.

    """
    if isinstance(obj, Pattern):
        ctx = context or {}

        def fn(node):
            return obj.match(node, ctx) is not NoMatch
    elif isinstance(obj, (tuple, type)):

        def fn(node):
            return isinstance(node, obj)
    elif callable(obj):
        fn = obj
    else:
        raise TypeError("finder must be callable, type, tuple of types or a pattern")

    return fn


def _coerce_replacer(obj: ReplacerLike, context: Optional[dict] = None) -> Replacer:
    """Coerce an object into a callable replacer function.

    Parameters
    ----------
    obj
        A Pattern, a Mapping or a callable which can be fed to `node.map()`
        to replace nodes.
    context
        Optional context to use if the replacer is a pattern.

    Returns
    -------
    A callable replacer function which can be used to replace nodes.

    """
    if isinstance(obj, Pattern):
        ctx = context or {}

        def fn(node, _, **kwargs):
            # need to first reconstruct the node from the possible rewritten
            # children, so we can match on the new node containing the rewritten
            # child arguments, this way we can propagate the rewritten nodes
            # upward in the hierarchy, using a specialized __recreate__ method
            # improves the performance by 17% compared node.__class__(**kwargs)
            recreated = node.__recreate__(kwargs)
            if (result := obj.match(recreated, ctx)) is NoMatch:
                return recreated
            else:
                return result

    elif isinstance(obj, Mapping):

        def fn(node, _, **kwargs):
            try:
                return obj[node]
            except KeyError:
                return node.__class__(**kwargs)
    elif callable(obj):
        fn = obj
    else:
        raise TypeError("replacer must be callable, mapping or a pattern")

    return fn


class Node(Hashable):
    __slots__ = ()

    @classmethod
    def __recreate__(cls, kwargs: Any) -> Self:
        """Reconstruct the node from the given arguments."""
        return cls(**kwargs)

    @property
    @abstractmethod
    def __args__(self) -> tuple[Any, ...]:
        """Sequence of arguments to traverse."""

    @property
    @abstractmethod
    def __argnames__(self) -> tuple[str, ...]:
        """Sequence of argument names."""

    @property
    def __children__(self) -> tuple[Node, ...]:
        """Sequence of children nodes."""
        return tuple(_flatten_collections(self.__args__))

    def __rich_repr__(self):
        """Support for rich reprerentation of the node."""
        return zip(self.__argnames__, self.__args__)

    def map(self, fn: Callable, filter: Optional[Finder] = None) -> dict[Node, Any]:
        """Apply a function to all nodes in the graph.

        The traversal is done in a topological order, so the function receives the
        results of its immediate children as keyword arguments.

        Parameters
        ----------
        fn
            Function to apply to each node. It receives the node as the first argument,
            the results as the second and the results of the children as keyword
            arguments.
        filter
            Pattern-like object to filter out nodes from the traversal. The traversal
            will only visit nodes that match the given pattern and stop otherwise.

        Returns
        -------
        A mapping of nodes to their results.

        """
        results: dict[Node, Any] = {}

        graph, _ = Graph.from_bfs(self, filter=filter).toposort()
        for node in graph:
            # minor optimization to directly recurse into the children
            kwargs = {
                k: _recursive_lookup(v, results)
                for k, v in zip(node.__argnames__, node.__args__)
            }
            results[node] = fn(node, results, **kwargs)

        return results

    @experimental
    def map_clear(
        self, fn: Callable, filter: Optional[Finder] = None
    ) -> dict[Node, Any]:
        """Apply a function to all nodes in the graph more memory efficiently.

        Alternative implementation of `map` to reduce memory usage. While `map` keeps
        all the results in memory until the end of the traversal, this method removes
        intermediate results as soon as they are not needed anymore.

        Prefer this method over `map` if the results consume significant amount of
        memory and if the intermediate results are not needed.

        Parameters
        ----------
        fn
            Function to apply to each node. It receives the node as the first argument,
            the results as the second and the results of the children as keyword
            arguments.
        filter
            Pattern-like object to filter out nodes from the traversal. The traversal
            will only visit nodes that match the given pattern and stop otherwise.

        Returns
        -------
        In contrast to `map`, this method returns the result of the root node only since
        the rest of the results are already discarded.

        """
        results: dict[Node, Any] = {}

        graph, dependents = Graph.from_bfs(self, filter=filter).toposort()
        dependents = {k: set(v) for k, v in dependents.items()}

        for node, dependencies in graph.items():
            kwargs = {
                k: _recursive_lookup(v, results)
                for k, v in zip(node.__argnames__, node.__args__)
            }
            results[node] = fn(node, results, **kwargs)

            # remove the results belonging to the dependencies if they are not
            # needed by other nodes during the rest of the traversal
            for dependency in set(dependencies):
                dependents[dependency].remove(node)
                if not dependents[dependency]:
                    del results[dependency]

        return results[self]

    # TODO(kszucs): perhaps rename it to find_all() for better clarity
    def find(
        self,
        finder: FinderLike,
        filter: Optional[FinderLike] = None,
        context: Optional[dict] = None,
    ) -> list[Node]:
        """Find all nodes matching a given pattern or type in the graph.

        Allow to match nodes based on the flexible pattern matching system implemented
        in the pattern module, but also provide a fast path for matching based on the
        type of the node.

        Parameters
        ----------
        finder
            A type, tuple of types, a pattern or a callable to match upon.
        filter
            A type, tuple of types, a pattern or a callable to filter out nodes
            from the traversal. The traversal will only visit nodes that match
            the given filter and stop otherwise.
        context
            Optional context to use if `finder` or `filter` is a pattern.

        Returns
        -------
        The list of nodes matching the given pattern. The order of the nodes is
        determined by a breadth-first search.

        """
        nodes = Graph.from_bfs(self, filter=filter, context=context).nodes()
        finder = _coerce_finder(finder, context)
        return [node for node in nodes if finder(node)]

    @experimental
    def find_topmost(
        self, finder: FinderLike, context: Optional[dict] = None
    ) -> list[Node]:
        """Find all topmost nodes matching a given pattern in the graph.

        A more advanced version of find, this method stops the traversal at the first
        node that matches the given pattern and does not descend into its children.

        Parameters
        ----------
        finder
            A type, tuple of types, a pattern or a callable to match upon.
        context
            Optional context to use if `finder` is a pattern.

        Returns
        -------
        The list of topmost nodes matching the given pattern.

        """
        seen = set()
        queue = deque([self])
        result = []
        finder = _coerce_finder(finder, context)

        while queue:
            if (node := queue.popleft()) not in seen:
                if finder(node):
                    result.append(node)
                else:
                    queue.extend(node.__children__)
                seen.add(node)
        return result

    @experimental
    def replace(
        self,
        replacer: ReplacerLike,
        filter: Optional[FinderLike] = None,
        context: Optional[dict] = None,
    ) -> Any:
        """Match and replace nodes in the graph according to a given pattern.

        The pattern matching system is used to match nodes in the graph and replace them
        with the results of the pattern.

        Parameters
        ----------
        replacer
            A `Pattern`, a `Mapping` or a callable which can be fed to
            `node.map()` directly to replace nodes.
        filter
            A type, tuple of types, a pattern or a callable to filter out nodes
            from the traversal. The traversal will only visit nodes that match
            the given filter and stop otherwise.
        context
            Optional context to use for the pattern matching.

        Returns
        -------
        The root node of the graph with the replaced nodes.

        """
        replacer = _coerce_replacer(replacer, context)
        results = self.map(replacer, filter=filter)
        return results.get(self, self)


class Graph(dict[Node, Sequence[Node]]):
    """A mapping-like graph data structure for easier graph traversal and manipulation.

    The data structure is a mapping of nodes to their children. The children are
    represented as a sequence of nodes. The graph can be constructed from a root node
    using the `from_bfs` or `from_dfs` class methods.

    Parameters
    ----------
    mapping : Node or Mapping[Node, Sequence[Node]], default ()
        Either a root node or a mapping of nodes to their children.

    """

    def __init__(self, mapping=(), /, **kwargs):
        if isinstance(mapping, Node):
            mapping = self.from_bfs(mapping)
        super().__init__(mapping, **kwargs)

    @classmethod
    def from_bfs(
        cls,
        root: Node,
        filter: Optional[FinderLike] = None,
        context: Optional[dict] = None,
    ) -> Self:
        """Construct a graph from a root node using a breadth-first search.

        The traversal is implemented in an iterative fashion using a queue.

        Parameters
        ----------
        root
            Root node of the graph.
        filter
            A type, tuple of types, a pattern or a callable to filter out nodes
            from the traversal. The traversal will only visit nodes that match
            the given filter and stop otherwise.
        context
            Optional context to use for the pattern matching.

        Returns
        -------
        A graph constructed from the root node.

        """
        if filter is None:
            return bfs(root)
        else:
            filter = _coerce_finder(filter, context)
            return bfs_while(root, filter=filter)

    @classmethod
    def from_dfs(
        cls,
        root: Node,
        filter: Optional[FinderLike] = None,
        context: Optional[dict] = None,
    ) -> Self:
        """Construct a graph from a root node using a depth-first search.

        The traversal is implemented in an iterative fashion using a stack.

        Parameters
        ----------
        root
            Root node of the graph.
        filter
            A type, tuple of types, a pattern or a callable to filter out nodes
            from the traversal. The traversal will only visit nodes that match
            the given filter and stop otherwise.
        context
            Optional context to use for the pattern matching.

        Returns
        -------
        A graph constructed from the root node.

        """
        if filter is None:
            return dfs(root)
        else:
            filter = _coerce_finder(filter, None)
            return dfs_while(root, filter=filter)

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()})"

    def nodes(self) -> KeysView[Node]:
        """Return all unique nodes in the graph."""
        return self.keys()

    def invert(self) -> Self:
        """Invert the data structure.

        The graph originally maps nodes to their children, this method inverts the
        mapping to map nodes to their parents.

        Returns
        -------
        The inverted graph.

        """
        result: dict[Node, list[Node]] = {node: [] for node in self}
        for node, dependencies in self.items():
            for dependency in dependencies:
                result[dependency].append(node)
        return self.__class__({k: tuple(v) for k, v in result.items()})

    def toposort(self) -> Self:
        """Topologically sort the graph using Kahn's algorithm.

        The graph is sorted in a way that all the dependencies of a node are placed
        before the node itself. The graph must not contain any cycles. Especially useful
        for mutating the graph in a way that the dependencies of a node are mutated
        before the node itself.

        Returns
        -------
        The topologically sorted graph.

        """
        dependents = self.invert()
        in_degree = {k: len(v) for k, v in self.items()}

        queue = deque(node for node, count in in_degree.items() if not count)
        result = self.__class__()

        while queue:
            node = queue.popleft()
            result[node] = self[node]

            for dependent in dependents[node]:
                in_degree[dependent] -= 1
                if not in_degree[dependent]:
                    queue.append(dependent)

        if any(in_degree.values()):
            raise ValueError("cycle detected in the graph")

        return result, dependents


# these could be callables instead
proceed = True
halt = False


def traverse(
    fn: Callable[[Node], tuple[bool | Iterable, Any]], node: Iterable[Node] | Node
) -> Iterator[Any]:
    """Utility for generic expression tree traversal.

    Parameters
    ----------
    fn
        A function applied on each expression. The first element of the tuple controls
        the traversal, and the second is the result if its not `None`.
    node
        The Node expression or a list of expressions.

    """

    args = reversed(node) if isinstance(node, Sequence) else [node]
    todo: deque[Node] = deque(args)
    seen: set[Node] = set()

    while todo:
        node = todo.pop()

        if node in seen:
            continue
        seen.add(node)

        control, result = fn(node)
        if result is not None:
            yield result

        if control is not halt:
            if control is proceed:
                children = node.__children__
            elif isinstance(control, Iterable):
                children = control
            else:
                raise TypeError(
                    "First item of the returned tuple must be "
                    "an instance of boolean or iterable"
                )

            todo.extend(reversed(children))


def bfs(root: Node) -> Graph:
    """Construct a graph from a root node using a breadth-first search.

    Parameters
    ----------
    root
        Root node of the graph.

    Returns
    -------
    A graph constructed from the root node.

    """
    # fast path for the default no filter case, according to benchmarks
    # this is gives a 10% speedup compared to the filtered version
    if not isinstance(root, Node):
        raise TypeError("node must be an instance of ibis.common.graph.Node")

    queue = deque([root])
    graph = Graph()

    while queue:
        if (node := queue.popleft()) not in graph:
            children = node.__children__
            graph[node] = children
            queue.extend(children)

    return graph


def bfs_while(root: Node, filter: Finder) -> Graph:
    """Construct a graph from a root node using a breadth-first search.

    Parameters
    ----------
    root
        Root node of the graph.
    filter
        A callable which returns a boolean given a node. The traversal will only
        visit nodes that match the given filter and stop otherwise.

    Returns
    -------
    A graph constructed from the root node.

    """
    if not isinstance(root, Node):
        raise TypeError("node must be an instance of ibis.common.graph.Node")

    queue = deque()
    graph = Graph()

    if filter(root):
        queue.append(root)

    while queue:
        if (node := queue.popleft()) not in graph:
            children = tuple(child for child in node.__children__ if filter(child))
            graph[node] = children
            queue.extend(children)

    return graph


def dfs(root: Node) -> Graph:
    """Construct a graph from a root node using a depth-first search.

    Parameters
    ----------
    root
        Root node of the graph.

    Returns
    -------
    A graph constructed from the root node.

    """
    # fast path for the default no filter case, according to benchmarks
    # this is gives a 10% speedup compared to the filtered version
    if not isinstance(root, Node):
        raise TypeError("node must be an instance of ibis.common.graph.Node")

    stack = deque([root])
    graph = {}

    while stack:
        if (node := stack.pop()) not in graph:
            children = node.__children__
            graph[node] = children
            stack.extend(children)

    return Graph(reversed(graph.items()))


def dfs_while(root: Node, filter: Finder) -> Graph:
    """Construct a graph from a root node using a depth-first search.

    Parameters
    ----------
    root
        Root node of the graph.
    filter
        A callable which returns a boolean given a node. The traversal will only
        visit nodes that match the given filter and stop otherwise.

    Returns
    -------
    A graph constructed from the root node.

    """
    if not isinstance(root, Node):
        raise TypeError("node must be an instance of ibis.common.graph.Node")

    stack = deque()
    graph = {}

    if filter(root):
        stack.append(root)

    while stack:
        if (node := stack.pop()) not in graph:
            children = tuple(child for child in node.__children__ if filter(child))
            graph[node] = children
            stack.extend(children)

    return Graph(reversed(graph.items()))



# Copyright 2015 Cloudera Inc
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import abc
import inspect

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import util

__all__ = ["scalar_function", "aggregate_function", "wrap_udf", "wrap_uda"]


class Function(abc.ABC):
    def __init__(self, inputs, output, name, database):
        self.inputs = tuple(map(dt.dtype, inputs))
        self.output = dt.dtype(output)
        self.name = name or util.guid()
        self.database = database
        self._klass = self._create_operation_class()

    @abc.abstractmethod
    def _create_operation_class(self):
        pass

    def __repr__(self):
        ident = ".".join(filter(None, (self.database, self.name)))
        return f"{ident}({self.inputs!r}, {self.output!r})"

    def __call__(self, *args):
        return self._klass(*args)

    def _make_fn(self):
        def fn(*args, **kwargs):
            ...

        fn.__name__ = self.name
        fn.__signature__ = inspect.Signature(
            parameters=[
                inspect.Parameter(
                    f"input{i:d}",
                    annotation=input,
                    kind=inspect.Parameter.POSITIONAL_ONLY,
                )
                for i, input in enumerate(self.inputs)
            ],
            return_annotation=self.output,
        )

        return fn


class ScalarFunction(Function):
    def _create_operation_class(self):
        return ops.scalar.builtin(
            fn=self._make_fn(),
            name=self.name,
            signature=(self.inputs, self.output),
            schema=self.database,
        )


class AggregateFunction(Function):
    def _create_operation_class(self):
        return ops.agg.builtin(
            fn=self._make_fn(),
            name=self.name,
            signature=(self.inputs, self.output),
            schema=self.database,
        )


class ImpalaFunction:
    def __init__(self, name=None, lib_path=None):
        self.lib_path = lib_path
        self.name = name or util.guid()

        if lib_path is not None:
            self._check_library()

    def _check_library(self):
        suffix = self.lib_path[-3:]
        if suffix not in [".so", ".ll"]:
            raise ValueError("Invalid file type. Must be .so or .ll ")


class ImpalaUDF(ScalarFunction, ImpalaFunction):
    """Feel free to customize my __doc__ or wrap in a nicer user API."""

    def __init__(
        self, inputs, output, so_symbol=None, lib_path=None, name=None, database=None
    ):
        from ibis.legacy.udf.validate import validate_output_type

        validate_output_type(output)
        self.so_symbol = so_symbol
        ImpalaFunction.__init__(self, name=name, lib_path=lib_path)
        ScalarFunction.__init__(self, inputs, output, name=self.name, database=database)


class ImpalaUDA(AggregateFunction, ImpalaFunction):
    def __init__(
        self,
        inputs,
        output,
        update_fn=None,
        init_fn=None,
        merge_fn=None,
        finalize_fn=None,
        serialize_fn=None,
        lib_path=None,
        name=None,
        database=None,
    ):
        self.init_fn = init_fn
        self.update_fn = update_fn
        self.merge_fn = merge_fn
        self.finalize_fn = finalize_fn
        self.serialize_fn = serialize_fn

        from ibis.legacy.udf.validate import validate_output_type

        validate_output_type(output)

        ImpalaFunction.__init__(self, name=name, lib_path=lib_path)
        AggregateFunction.__init__(
            self, inputs, output, name=self.name, database=database
        )

    def _check_library(self):
        suffix = self.lib_path[-3:]
        if suffix == ".ll":
            raise com.IbisInputError("LLVM IR UDAs are not yet supported")
        elif suffix != ".so":
            raise ValueError("Invalid file type. Must be .so")


def wrap_uda(
    hdfs_file: str,
    inputs: str,
    output: str,
    update_fn: str,
    init_fn: str | None = None,
    merge_fn: str | None = None,
    finalize_fn: str | None = None,
    serialize_fn: str | None = None,
    name: str | None = None,
    database: str | None = None,
):
    """Creates a callable aggregation function object.

    Must be created in Impala to be used.

    Parameters
    ----------
    hdfs_file
        .so file that contains relevant UDA
    inputs
        list of strings denoting ibis datatypes
    output
        string denoting ibis datatype
    update_fn
        Library symbol name for update function
    init_fn
        Library symbol name for initialization function
    merge_fn
        Library symbol name for merge function
    finalize_fn
        Library symbol name for finalize function
    serialize_fn
        Library symbol name for serialize UDA API function. Not required for all
        UDAs.
    name
        Used internally to track function
    database
        Name of database

    """
    return ImpalaUDA(
        inputs,
        output,
        update_fn,
        init_fn,
        merge_fn,
        finalize_fn,
        serialize_fn=serialize_fn,
        name=name,
        lib_path=hdfs_file,
        database=database,
    )


def wrap_udf(hdfs_file, inputs, output, so_symbol, name=None, database=None):
    """Creates a callable scalar function object.

    Must be created in Impala to be used.

    Parameters
    ----------
    hdfs_file
        .so file that contains relevant UDF
    inputs
        Input types to UDF
    output
        Ibis data type
    so_symbol
        C++ function name for relevant UDF
    name
        Used internally to track function
    database
        Name of database

    """
    func = ImpalaUDF(
        inputs, output, so_symbol, name=name, lib_path=hdfs_file, database=database
    )
    return func


def scalar_function(inputs, output, name=None, database=None):
    """Create an operator class.

    Parameters
    ----------
    inputs
        Ibis data type names
    output
        Ibis data type
    name
        Used internally to track function
    database
        Name of database

    """
    return ScalarFunction(inputs, output, name=name, database=database)


def aggregate_function(inputs, output, name=None, database=None):
    """Create an operator class.

    Parameters
    ----------
    inputs
        Ibis data type names
    output
        Ibis data type
    name
        Used internally to track function
    database
        Name of database

    """
    return AggregateFunction(inputs, output, name=name, database=database)



import ibis


employee = ibis.table(
    name="employee",
    schema={"first_name": "string", "last_name": "string", "id": "int64"},
)

result = employee.aggregate([employee.first_name.count().name("_col_0")])



from __future__ import annotations

import pandas.testing as tm
import pytest

import ibis.expr.datatypes as dt
import ibis.expr.types as ir
from ibis import udf
from ibis.legacy.udf.vectorized import elementwise, reduction

pytest.importorskip("datafusion")
pc = pytest.importorskip("pyarrow.compute")


@elementwise(input_type=["string"], output_type="int64")
def my_string_length(arr, **kwargs):
    # arr is a pyarrow.StringArray
    return pc.cast(pc.multiply(pc.utf8_length(arr), 2), target_type="int64")


@elementwise(input_type=[dt.int64, dt.int64], output_type=dt.int64)
def my_add(arr1, arr2, **kwargs):
    return pc.add(arr1, arr2)


@reduction(input_type=[dt.float64], output_type=dt.float64)
def my_mean(arr):
    return pc.mean(arr)


def test_udf(alltypes):
    data_string_col = alltypes.date_string_col.execute()
    expected = data_string_col.str.len() * 2

    expr = my_string_length(alltypes.date_string_col)
    assert isinstance(expr, ir.Column)

    result = expr.execute()
    tm.assert_series_equal(result, expected, check_names=False)


def test_multiple_argument_udf(alltypes):
    expr = my_add(alltypes.smallint_col, alltypes.int_col).name("tmp")
    result = expr.execute()

    df = alltypes[["smallint_col", "int_col"]].execute()
    expected = (df.smallint_col + df.int_col).astype("int64")

    tm.assert_series_equal(result, expected.rename("tmp"))


def test_builtin_scalar_udf(con):
    @udf.scalar.builtin
    def to_hex(a: int) -> str:
        """Convert an integer to a hex string."""

    expr = to_hex(42)
    result = con.execute(expr)
    assert result == "2a"


def test_builtin_agg_udf(con):
    @udf.agg.builtin
    def median(a: float) -> float:
        """Median of a column."""

    expr = median(con.tables.batting.G)
    result = con.execute(expr)
    assert result == con.tables.batting.G.execute().median()


def test_builtin_agg_udf_filtered(con):
    @udf.agg.builtin
    def median(a: float, where: bool = True) -> float:
        """Median of a column."""

    median(con.tables.batting.G).execute()



from __future__ import annotations

import pytest

import ibis

catalog = {
    "employee": {"first_name": "string", "last_name": "string", "id": "int64"},
    "call": {
        "start_time": "timestamp",
        "end_time": "timestamp",
        "employee_id": "int64",
        "call_outcome_id": "int64",
        "call_attempts": "int64",
    },
    "call_outcome": {"outcome_text": "string", "id": "int64"},
}


def test_parse_sql_basic_projection(snapshot):
    sql = "SELECT *, first_name as first FROM employee WHERE id < 5 ORDER BY id DESC"
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


@pytest.mark.parametrize("how", ["right", "left", "inner"])
def test_parse_sql_basic_join(how, snapshot):
    sql = f"""
SELECT
  *,
  first_name as first
FROM employee {how.upper()}
JOIN call ON
  employee.id = call.employee_id
WHERE
  id < 5
ORDER BY
  id DESC"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_multiple_joins(snapshot):
    sql = """
SELECT *
FROM employee
JOIN call
  ON employee.id = call.employee_id
JOIN call_outcome
  ON call.call_outcome_id = call_outcome.id"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_basic_aggregation(snapshot):
    sql = """
SELECT
  employee_id,
  sum(call_attempts) AS attempts
FROM call
GROUP BY employee_id"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_basic_aggregation_with_join(snapshot):
    sql = """
SELECT
  id,
  sum(call_attempts) AS attempts
FROM employee
LEFT JOIN call
  ON employee.id = call.employee_id
GROUP BY id"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_aggregation_with_multiple_joins(snapshot):
    sql = """
SELECT
  t.employee_id,
  AVG(t.call_attempts) AS avg_attempts
FROM (
  SELECT * FROM employee JOIN call ON employee.id = call.employee_id
  JOIN call_outcome ON call.call_outcome_id = call_outcome.id
) AS t
GROUP BY t.employee_id"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_simple_reduction(snapshot):
    sql = """SELECT AVG(call_attempts) AS mean FROM call"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_scalar_subquery(snapshot):
    sql = """
SELECT *
FROM call
WHERE call_attempts > (
  SELECT avg(call_attempts) AS mean
  FROM call
)"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_simple_select_count(snapshot):
    sql = """SELECT COUNT(first_name) FROM employee"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_table_alias(snapshot):
    sql = """SELECT e.* FROM employee AS e"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_join_with_filter(snapshot):
    sql = """
SELECT *, first_name as first FROM employee
LEFT JOIN call ON employee.id = call.employee_id
WHERE id < 5
ORDER BY id DESC"""
    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")


def test_parse_sql_in_clause(snapshot):
    sql = """
SELECT first_name FROM employee
WHERE first_name IN ('Graham', 'John', 'Terry', 'Eric', 'Michael')"""

    expr = ibis.parse_sql(sql, catalog)
    code = ibis.decompile(expr, format=True)
    snapshot.assert_match(code, "decompiled.py")



from __future__ import annotations

import contextlib
import csv
import io
import sqlite3
from typing import Any

import pytest

import ibis
import ibis.expr.types as ir
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.base import BackendTest


class TestConf(BackendTest):
    supports_arrays = False
    supports_arrays_outside_of_select = supports_arrays
    supports_window_operations = True
    check_dtype = False
    returned_timestamp_unit = "s"
    supports_structs = False
    stateful = False
    deps = ("regex",)

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw):
        return ibis.sqlite.connect(**kw)

    def _load_data(self, **kw: Any) -> None:
        """Load test data into a SQLite backend instance."""
        with self.connection.begin() as con:
            for stmt in self.ddl_script:
                con.execute(stmt)

        with self.connection.begin() as con:
            for table in TEST_TABLES:
                basename = f"{table}.csv"
                with self.data_dir.joinpath("csv", basename).open(
                    "r", encoding="UTF-8"
                ) as f:
                    if basename == "astronauts.csv":
                        input = io.StringIO(f.read().replace("\n ", " "))
                    else:
                        input = f
                    reader = csv.reader(input)
                    header = next(reader)
                    assert header, f"empty header for table: `{table}`"
                    spec = ", ".join("?" * len(header))
                    with contextlib.closing(con.connection.cursor()) as cur:
                        cur.executemany(f"INSERT INTO {table} VALUES ({spec})", reader)

    @property
    def functional_alltypes(self) -> ir.Table:
        t = super().functional_alltypes
        return t.mutate(timestamp_col=t.timestamp_col.cast("timestamp"))


@pytest.fixture
def dbpath(tmp_path):
    path = tmp_path / "test.db"
    con = sqlite3.connect(path)
    con.execute("CREATE TABLE t AS SELECT 1 a UNION SELECT 2 UNION SELECT 3")
    con.execute("CREATE TABLE s AS SELECT 1 b UNION SELECT 2")
    return path


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection


@pytest.fixture(scope="session")
def translate(dialect):
    return lambda expr: ibis.to_sql(expr, dialect="sqlite")


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture(scope="session")
def df(alltypes):
    return alltypes.execute()



from __future__ import annotations

import ibis
from ibis import _


def test_window_with_row_number_compiles():
    # GH #8058: the add_order_by_to_empty_ranking_window_functions rule was
    # matching on `RankBase` subclasses with a pattern expecting an `arg`
    # attribute, which is not present on `RowNumber`
    expr = (
        ibis.memtable({"a": range(30)})
        .mutate(id=ibis.row_number())
        .sample(fraction=0.25, seed=0)
        .mutate(is_test=_.id.isin(_.id))
        .filter(~_.is_test)
    )
    assert ibis.to_sql(expr)



from __future__ import annotations

import string

import hypothesis as h
import hypothesis.strategies as st
import parsy
import pytest

import ibis.expr.datatypes as dt
import ibis.tests.strategies as its
from ibis.common.annotations import ValidationError


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ("boolean", dt.boolean),
        ("int8", dt.int8),
        ("int16", dt.int16),
        ("int32", dt.int32),
        ("int64", dt.int64),
        ("int", dt.int64),
        ("uint8", dt.uint8),
        ("uint16", dt.uint16),
        ("uint32", dt.uint32),
        ("uint64", dt.uint64),
        ("float16", dt.float16),
        ("float32", dt.float32),
        ("float64", dt.float64),
        ("float", dt.float64),
        ("string", dt.string),
        ("binary", dt.binary),
        ("date", dt.date),
        ("time", dt.time),
        ("timestamp", dt.timestamp),
        ("point", dt.point),
        ("linestring", dt.linestring),
        ("polygon", dt.polygon),
        ("multilinestring", dt.multilinestring),
        ("multipoint", dt.multipoint),
        ("multipolygon", dt.multipolygon),
    ],
)
def test_primitive_from_string(spec, expected):
    assert dt.dtype(spec) == expected


@pytest.mark.parametrize(
    ("spec", "expected"),
    [
        ["decimal", dt.Decimal(None, None)],
        ["decimal(10, 3)", dt.Decimal(10, 3)],
        ["bignumeric", dt.Decimal(76, 38)],
        ["bigdecimal", dt.Decimal(76, 38)],
        ["bignumeric(1, 1)", dt.Decimal(1, 1)],
        ["bigdecimal(1, 1)", dt.Decimal(1, 1)],
    ],
)
def test_parse_decimal(spec, expected):
    assert dt.dtype(spec) == expected


@pytest.mark.parametrize(
    "case",
    [
        "decimal(",
        "decimal()",
        "decimal(3)",
        "decimal(,)",
        "decimal(3,)",
        "decimal(3,",
    ],
)
def test_parse_decimal_failure(case):
    with pytest.raises(parsy.ParseError):
        dt.dtype(case)


@pytest.mark.parametrize("spec", ["varchar", "varchar(10)", "char", "char(10)"])
def test_parse_char_varchar(spec):
    assert dt.dtype(spec) == dt.string


@pytest.mark.parametrize(
    "spec", ["varchar(", "varchar)", "varchar()", "char(", "char)", "char()"]
)
def test_parse_char_varchar_invalid(spec):
    with pytest.raises(parsy.ParseError):
        dt.dtype(spec)


def test_parse_array_token_error():
    with pytest.raises(parsy.ParseError):
        dt.dtype("array<string>>")


def test_parse_struct():
    orders = """array<struct<
                    oid: int64,
                    status: string,
                    totalprice: decimal(12, 2),
                    order_date: string,
                    items: array<struct<
                        iid: int64,
                        name: string,
                        price: decimal(12, 2),
                        discount_perc: decimal(12, 2),
                        shipdate: string,
                        : bool
                    >>
                >>"""
    expected = dt.Array(
        dt.Struct.from_tuples(
            [
                ("oid", dt.int64),
                ("status", dt.string),
                ("totalprice", dt.Decimal(12, 2)),
                ("order_date", dt.string),
                (
                    "items",
                    dt.Array(
                        dt.Struct.from_tuples(
                            [
                                ("iid", dt.int64),
                                ("name", dt.string),
                                ("price", dt.Decimal(12, 2)),
                                ("discount_perc", dt.Decimal(12, 2)),
                                ("shipdate", dt.string),
                                ("", dt.boolean),
                            ]
                        )
                    ),
                ),
            ]
        )
    )

    assert dt.dtype(orders) == expected


def test_struct_with_string_types():
    result = dt.Struct.from_tuples(
        [
            ("a", "map<double, string>"),
            ("b", "array<map<string, array<int32>>>"),
            ("c", "array<string>"),
            ("d", "int8"),
        ]
    )

    assert result == dt.Struct.from_tuples(
        [
            ("a", dt.Map(dt.double, dt.string)),
            ("b", dt.Array(dt.Map(dt.string, dt.Array(dt.int32)))),
            ("c", dt.Array(dt.string)),
            ("d", dt.int8),
        ]
    )


def test_array_with_string_value_types():
    assert dt.Array("int32") == dt.Array(dt.int32)
    assert dt.Array(dt.Array("array<map<string, double>>")) == (
        dt.Array(dt.Array(dt.Array(dt.Map(dt.string, dt.double))))
    )


def test_map_with_string_value_types():
    assert dt.Map("int32", "double") == dt.Map(dt.int32, dt.double)
    assert dt.Map("int32", "array<double>") == dt.Map(dt.int32, dt.Array(dt.double))


def test_parse_empty_map_failure():
    with pytest.raises(parsy.ParseError):
        dt.dtype("map<>")


def test_parse_map_allow_non_primitive_keys():
    assert dt.dtype("map<array<string>, double>") == dt.Map(
        dt.Array(dt.string), dt.double
    )


def test_parse_timestamp_with_timezone_single_quote():
    t = dt.dtype("timestamp('US/Eastern')")
    assert isinstance(t, dt.Timestamp)
    assert t.timezone == "US/Eastern"


def test_parse_timestamp_with_timezone_double_quote():
    t = dt.dtype('timestamp("US/Eastern")')
    assert isinstance(t, dt.Timestamp)
    assert t.timezone == "US/Eastern"


def test_parse_timestamp_with_timezone_invalid_timezone():
    ts = dt.dtype("timestamp('US/Ea')")
    assert str(ts) == "timestamp('US/Ea')"


@pytest.mark.parametrize("scale", range(10))
@pytest.mark.parametrize("tz", ["UTC", "America/New_York"])
def test_parse_timestamp_with_scale(scale, tz):
    expected = dt.Timestamp(timezone=tz, scale=scale)
    typestring = f"timestamp({tz!r}, {scale:d})"
    assert dt.parse(typestring) == expected
    assert str(expected) == typestring


@pytest.mark.parametrize("scale", range(10))
def test_parse_timestamp_with_scale_no_tz(scale):
    assert dt.parse(f"timestamp({scale:d})") == dt.Timestamp(scale=scale)


@pytest.mark.parametrize(
    "unit",
    [
        "Y",
        "Q",
        "M",
        "W",
        "D",  # date units
        "h",
        "m",
        "s",
        "ms",
        "us",
        "ns",  # time units
    ],
)
def test_parse_interval(unit):
    definition = f"interval('{unit}')"
    assert dt.Interval(unit) == dt.dtype(definition)


@pytest.mark.parametrize("unit", ["X", "unsupported"])
def test_parse_interval_with_invalid_unit(unit):
    definition = f"interval('{unit}')"
    with pytest.raises(ValidationError):
        dt.dtype(definition)


@pytest.mark.parametrize(
    "case",
    [
        "timestamp(US/Ea)",
        "timestamp('US/Eastern\")",
        "timestamp(\"US/Eastern')",
        "interval(Y)",
        "interval('Y\")",
        "interval(\"Y')",
    ],
)
def test_parse_temporal_with_invalid_string_argument(case):
    with pytest.raises(parsy.ParseError):
        dt.dtype(case)


def test_parse_time():
    assert dt.dtype("time").equals(dt.time)


def test_parse_null():
    assert dt.parse("null") == dt.null


# corresponds to its.all_dtypes() but without:
# - geospacial types, the string representation is different from what the parser expects
# - struct types, the generated struct field names contain special characters

field_names = st.text(
    alphabet=st.characters(
        whitelist_characters=string.ascii_letters + string.digits,
        whitelist_categories=(),
    )
)

roundtrippable_dtypes = st.deferred(
    lambda: (
        its.primitive_dtypes()
        | its.string_like_dtypes()
        | its.temporal_dtypes()
        | its.interval_dtype()
        | its.variadic_dtypes()
        | its.struct_dtypes(names=field_names)
        | its.array_dtypes(roundtrippable_dtypes)
        | its.map_dtypes(roundtrippable_dtypes, roundtrippable_dtypes)
    )
)


@h.given(roundtrippable_dtypes)
def test_parse_dtype_roundtrip(dtype):
    assert dt.dtype(str(dtype)) == dtype



import ibis


call = ibis.table(
    name="call",
    schema={
        "start_time": "timestamp",
        "end_time": "timestamp",
        "employee_id": "int64",
        "call_outcome_id": "int64",
        "call_attempts": "int64",
    },
)
agg = call.aggregate([call.call_attempts.mean().name("mean")])

result = call.inner_join(
    agg, [agg.mean < call.call_attempts, ibis.literal(True)]
).select(
    call.start_time,
    call.end_time,
    call.employee_id,
    call.call_outcome_id,
    call.call_attempts,
    agg.mean,
)



from __future__ import annotations

import sqlglot as sg
import sqlglot.expressions as sge
import toolz

import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis.backends.base.sqlglot.compiler import NULL, SQLGlotCompiler
from ibis.backends.base.sqlglot.datatypes import DruidType
from ibis.backends.base.sqlglot.dialects import Druid
from ibis.backends.base.sqlglot.rewrites import (
    rewrite_capitalize,
    rewrite_sample_as_filter,
)


class DruidCompiler(SQLGlotCompiler):
    __slots__ = ()

    dialect = Druid
    type_mapper = DruidType
    rewrites = (
        rewrite_sample_as_filter,
        *(
            rewrite
            for rewrite in SQLGlotCompiler.rewrites
            if rewrite is not rewrite_capitalize
        ),
    )

    UNSUPPORTED_OPERATIONS = frozenset(
        (
            ops.ApproxMedian,
            ops.Arbitrary,
            ops.ArgMax,
            ops.ArgMin,
            ops.ArrayCollect,
            ops.ArrayDistinct,
            ops.ArrayFilter,
            ops.ArrayFlatten,
            ops.ArrayIntersect,
            ops.ArrayMap,
            ops.ArraySort,
            ops.ArrayUnion,
            ops.ArrayZip,
            ops.CountDistinctStar,
            ops.Covariance,
            ops.DateDelta,
            ops.DayOfWeekIndex,
            ops.DayOfWeekName,
            ops.First,
            ops.IntervalFromInteger,
            ops.IsNan,
            ops.IsInf,
            ops.Last,
            ops.Levenshtein,
            ops.Median,
            ops.MultiQuantile,
            ops.Quantile,
            ops.RegexReplace,
            ops.RegexSplit,
            ops.RowID,
            ops.StandardDev,
            ops.Strftime,
            ops.StringAscii,
            ops.StringSplit,
            ops.StringToTimestamp,
            ops.TimeDelta,
            ops.TimestampBucket,
            ops.TimestampDelta,
            ops.TimestampNow,
            ops.Translate,
            ops.TypeOf,
            ops.Unnest,
            ops.Variance,
        )
    )

    SIMPLE_OPS = {
        ops.BitAnd: "bit_and",
        ops.BitOr: "bit_or",
        ops.BitXor: "bit_xor",
        ops.BitwiseAnd: "bitwise_and",
        ops.BitwiseNot: "bitwise_complement",
        ops.BitwiseOr: "bitwise_or",
        ops.BitwiseXor: "bitwise_xor",
        ops.BitwiseLeftShift: "bitwise_shift_left",
        ops.BitwiseRightShift: "bitwise_shift_right",
        ops.Modulus: "mod",
        ops.Power: "power",
        ops.Log10: "log10",
        ops.ApproxCountDistinct: "approx_count_distinct",
        ops.StringContains: "contains_string",
    }

    def _aggregate(self, funcname: str, *args, where):
        expr = self.f[funcname](*args)
        if where is not None:
            return sg.exp.Filter(this=expr, expression=sg.exp.Where(this=where))
        return expr

    def visit_Sum(self, op, *, arg, where):
        arg = self.if_(arg, 1, 0) if op.arg.dtype.is_boolean() else arg
        return self.agg.sum(arg, where=where)

    def visit_InMemoryTable(self, op, *, name, schema, data):
        # the performance of this is rather terrible
        tuples = data.to_frame().itertuples(index=False)
        quoted = self.quoted
        columns = [sg.column(col, quoted=quoted) for col in schema.names]
        expr = sge.Values(
            expressions=[
                sge.Tuple(expressions=tuple(map(sge.convert, row))) for row in tuples
            ],
            alias=sge.TableAlias(
                this=sg.to_identifier(name, quoted=quoted),
                columns=columns,
            ),
        )
        return sg.select(*columns).from_(expr)

    def visit_StringJoin(self, op, *, arg, sep):
        return self.f.concat(*toolz.interpose(sep, arg))

    def visit_Pi(self, op):
        return self.f.acos(-1)

    def visit_Sign(self, op, *, arg):
        return self.if_(arg.eq(0), 0, self.if_(arg > 0, 1, -1))

    def visit_GroupConcat(self, op, *, arg, sep, where):
        return self.agg.string_agg(arg, sep, 1 << 20, where=where)

    def visit_StartsWith(self, op, *, arg, start):
        return self.f.left(arg, self.f.length(start)).eq(start)

    def visit_EndsWith(self, op, *, arg, end):
        return self.f.right(arg, self.f.length(end)).eq(end)

    def visit_Capitalize(self, op, *, arg):
        return self.if_(
            self.f.length(arg) < 2,
            self.f.upper(arg),
            self.f.concat(
                self.f.upper(self.f.substr(arg, 1, 1)),
                self.f.lower(self.f.substr(arg, 2)),
            ),
        )

    def visit_RegexSearch(self, op, *, arg, pattern):
        return self.f.anon.regexp_like(arg, pattern)

    def visit_StringSQLILike(self, op, *, arg, pattern, escape):
        if escape is not None:
            raise NotImplementedError("non-None escape not supported")
        return self.f.upper(arg).like(self.f.upper(pattern))

    def visit_Literal(self, op, *, value, dtype):
        if value is None:
            return NULL
        return super().visit_Literal(op, value=value, dtype=dtype)

    def visit_NonNullLiteral(self, op, *, value, dtype):
        if dtype.is_uuid():
            return sge.convert(str(value))

        return None

    def visit_Cast(self, op, *, arg, to):
        from_ = op.arg.dtype
        if from_.is_integer() and to.is_timestamp():
            # seconds since UNIX epoch
            return self.f.millis_to_timestamp(arg * 1_000)
        elif from_.is_string() and to.is_timestamp():
            return self.f.time_parse(arg)
        return super().visit_Cast(op, arg=arg, to=to)

    def visit_TimestampFromYMDHMS(
        self, op, *, year, month, day, hours, minutes, seconds
    ):
        return self.f.time_parse(
            self.f.concat(
                self.f.lpad(self.cast(year, dt.string), 4, "0"),
                "-",
                self.f.lpad(self.cast(month, dt.string), 2, "0"),
                "-",
                self.f.lpad(self.cast(day, dt.string), 2, "0"),
                "T",
                self.f.lpad(self.cast(hours, dt.string), 2, "0"),
                ":",
                self.f.lpad(self.cast(minutes, dt.string), 2, "0"),
                ":",
                self.f.lpad(self.cast(seconds, dt.string), 2, "0"),
                "Z",
            )
        )



from __future__ import annotations

import os
import tempfile
from pathlib import Path

import pandas as pd
import pandas.testing as tm
import pyarrow as pa
import pytest

import ibis
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.tests.errors import Py4JJavaError


@pytest.fixture
def tempdir_sink_configs():
    def generate_tempdir_configs(tempdir):
        return {"connector": "filesystem", "path": tempdir, "format": "csv"}

    return generate_tempdir_configs


@pytest.mark.parametrize("temp", [True, False])
def test_list_tables(con, temp):
    assert len(con.list_tables(temp=temp))
    assert con.list_tables(catalog="default_catalog", database="default_database")


def test_create_table_from_schema(
    con, awards_players_schema, temp_table, csv_source_configs
):
    new_table = con.create_table(
        temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
    )
    assert temp_table in con.list_tables()
    assert new_table.schema() == awards_players_schema


@pytest.mark.parametrize("temp", [True, False])
def test_create_table(con, awards_players_schema, temp_table, csv_source_configs, temp):
    con.create_table(
        temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
        temp=temp,
    )
    assert temp_table in con.list_tables()

    if temp:
        with pytest.raises(Py4JJavaError):
            con.drop_table(temp_table)

    con.drop_table(temp_table, temp=temp)

    assert temp_table not in con.list_tables()


def test_recreate_table_from_schema(
    con, awards_players_schema, temp_table, csv_source_configs
):
    # create table once
    new_table = con.create_table(
        temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
    )
    assert temp_table in con.list_tables()
    assert new_table.schema() == awards_players_schema

    # create the same table a second time should fail
    with pytest.raises(
        Py4JJavaError,
        match="org.apache.flink.table.catalog.exceptions.TableAlreadyExistException",
    ):
        new_table = con.create_table(
            temp_table,
            schema=awards_players_schema,
            tbl_properties=csv_source_configs("awards_players"),
            overwrite=False,
        )


def test_force_recreate_table_from_schema(
    con, awards_players_schema, temp_table, csv_source_configs
):
    # create table once
    new_table = con.create_table(
        temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
    )
    assert temp_table in con.list_tables()
    assert new_table.schema() == awards_players_schema

    # force creating the same twice a second time
    new_table = con.create_table(
        temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
        overwrite=True,
    )
    assert temp_table in con.list_tables()
    assert new_table.schema() == awards_players_schema


@pytest.mark.parametrize(
    "schema, table_name",
    [(None, None), (TEST_TABLES["awards_players"], "awards_players")],
)
def test_recreate_in_mem_table(con, schema, table_name, temp_table, csv_source_configs):
    employee_df = pd.DataFrame(
        [("fred flintstone", "award", 2002, "lg_id", "tie", "this is a note")]
    )
    # create table once
    if table_name is not None:
        tbl_properties = csv_source_configs(table_name)
    else:
        tbl_properties = None

    new_table = con.create_table(
        name=temp_table,
        obj=employee_df,
        schema=schema,
        tbl_properties=tbl_properties,
        temp=True,
    )
    try:
        assert temp_table in con.list_tables()
        if schema is not None:
            assert new_table.schema() == schema

        # create the same table a second time should fail
        with pytest.raises(
            Py4JJavaError,
            match=r"An error occurred while calling o\d+\.createTemporaryView",
        ):
            new_table = con.create_table(
                name=temp_table,
                obj=employee_df,
                schema=schema,
                tbl_properties=tbl_properties,
                overwrite=False,
                temp=True,
            )
    finally:
        con.drop_table(temp_table, force=True)


@pytest.mark.parametrize(
    "schema_props", [(None, None), (TEST_TABLES["awards_players"], "awards_players")]
)
def test_force_recreate_in_mem_table(con, schema_props, temp_table, csv_source_configs):
    employee_df = pd.DataFrame(
        [("fred flintstone", "award", 2002, "lg_id", "tie", "this is a note")]
    )
    # create table once
    schema = schema_props[0]
    if schema_props[1] is not None:
        tbl_properties = csv_source_configs(schema_props[1])
    else:
        tbl_properties = None

    new_table = con.create_table(
        name=temp_table,
        obj=employee_df,
        schema=schema,
        tbl_properties=tbl_properties,
        temp=True,
    )
    try:
        assert temp_table in con.list_tables()
        if schema is not None:
            assert new_table.schema() == schema

        # force recreate the same table a second time should succeed
        new_table = con.create_table(
            name=temp_table,
            obj=employee_df,
            schema=schema,
            tbl_properties=tbl_properties,
            temp=True,
            overwrite=True,
        )
        assert temp_table in con.list_tables()
        if schema is not None:
            assert new_table.schema() == schema
    finally:
        con.drop_table(temp_table, force=True)


@pytest.fixture
def functional_alltypes_schema_w_nonnullable_columns():
    return sch.Schema(
        {
            "id": dt.int32(nullable=False),
            "bool_col": dt.bool(nullable=False),
            "smallint_col": dt.int16(nullable=False),
            "int_col": dt.int32(nullable=False),
            "bigint_col": dt.int64(nullable=False),
            "float_col": dt.float32(nullable=False),
            "double_col": dt.float64(nullable=False),
            "date_string_col": dt.string(nullable=False),
            "string_col": dt.string(nullable=False),
            "year": dt.int32(nullable=False),
            "month": dt.int32(nullable=False),
            "timestamp_col": dt.timestamp(scale=3),
        }
    )


@pytest.mark.parametrize(
    "primary_key",
    [
        None,
        "id",
        ["id"],
        ["month"],
        ["id", "string_col"],
        ["id", "string_col", "year"],
    ],
)
def test_create_source_table_with_watermark_and_primary_key(
    con,
    temp_table,
    functional_alltypes_schema_w_nonnullable_columns,
    csv_source_configs,
    primary_key,
):
    new_table = con.create_table(
        temp_table,
        schema=functional_alltypes_schema_w_nonnullable_columns,
        tbl_properties=csv_source_configs("functional_alltypes"),
        watermark=ibis.watermark(
            time_col="timestamp_col", allowed_delay=ibis.interval(seconds=15)
        ),
        primary_key=primary_key,
    )
    try:
        assert temp_table in con.list_tables()
        assert new_table.schema() == functional_alltypes_schema_w_nonnullable_columns
    finally:
        con.drop_table(temp_table, force=True)


@pytest.mark.parametrize(
    "primary_key",
    [
        "nonexistent_column",
        ["nonexistent_column"],
        ["id", "nonexistent_column"],
    ],
)
def test_create_table_failure_with_invalid_primary_keys(
    con,
    temp_table,
    functional_alltypes_schema_w_nonnullable_columns,
    csv_source_configs,
    primary_key,
):
    with pytest.raises(exc.IbisError):
        con.create_table(
            temp_table,
            schema=functional_alltypes_schema_w_nonnullable_columns,
            tbl_properties=csv_source_configs("functional_alltypes"),
            primary_key=primary_key,
        )
    assert temp_table not in con.list_tables()


@pytest.fixture
def temp_view(con):
    name = ibis.util.gen_name("view")
    yield name
    con.drop_view(name, force=True)


@pytest.mark.parametrize("temp", [True, False])
def test_create_view(
    con, temp_table, awards_players_schema, csv_source_configs, temp_view, temp
):
    table = con.create_table(
        name=temp_table,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
    )
    assert temp_table in con.list_tables()

    con.create_view(
        name=temp_view,
        obj=table,
        force=False,
        temp=temp,
        overwrite=False,
    )
    view_list = sorted(con.list_tables())
    assert temp_view in view_list

    # Try to re-create the same view with `force=False`
    with pytest.raises(Py4JJavaError):
        con.create_view(
            name=temp_view,
            obj=table,
            force=False,
            temp=temp,
            overwrite=False,
        )
    assert view_list == sorted(con.list_tables())

    # Try to re-create the same view with `force=True`
    con.create_view(
        name=temp_view,
        obj=table,
        force=True,
        temp=temp,
        overwrite=False,
    )
    assert view_list == sorted(con.list_tables())

    # Overwrite the view
    con.create_view(
        name=temp_view,
        obj=table,
        force=False,
        temp=temp,
        overwrite=True,
    )
    assert view_list == sorted(con.list_tables())

    con.drop_view(name=temp_view, temp=temp, force=True)
    assert temp_view not in con.list_tables()


def test_rename_table(con, awards_players_schema, temp_table, csv_source_configs):
    table_name = temp_table
    con.create_table(
        name=table_name,
        schema=awards_players_schema,
        tbl_properties=csv_source_configs("awards_players"),
    )
    assert table_name in con.list_tables()

    new_table_name = f"{table_name}_new"
    con.rename_table(
        old_name=table_name,
        new_name=new_table_name,
        force=False,
    )
    assert table_name not in con.list_tables()
    assert new_table_name in con.list_tables()

    con.drop_table(new_table_name)
    assert new_table_name not in con.list_tables()


@pytest.mark.parametrize(
    "obj",
    [
        pytest.param(
            [("fred flintstone", 35, 1.28), ("barney rubble", 32, 2.32)], id="list"
        ),
        pytest.param(
            {
                "name": ["fred flintstone", "barney rubble"],
                "age": [35, 32],
                "gpa": [1.28, 2.32],
            },
            id="dict",
        ),
        pytest.param(
            pd.DataFrame(
                [("fred flintstone", 35, 1.28), ("barney rubble", 32, 2.32)],
                columns=["name", "age", "gpa"],
            ),
            id="pandas_dataframe",
        ),
        pytest.param(
            pa.Table.from_arrays(
                [
                    pa.array(["fred flintstone", "barney rubble"]),
                    pa.array([35, 32]),
                    pa.array([1.28, 2.32]),
                ],
                names=["name", "age", "gpa"],
            ),
            id="pyarrow_table",
        ),
    ],
)
def test_insert_values_into_table(con, tempdir_sink_configs, obj):
    sink_schema = sch.Schema({"name": dt.string, "age": dt.int64, "gpa": dt.float64})
    with tempfile.TemporaryDirectory() as tempdir:
        con.create_table(
            "tempdir_sink",
            schema=sink_schema,
            tbl_properties=tempdir_sink_configs(tempdir),
        )
        try:
            con.insert("tempdir_sink", obj).wait()
            temporary_file = next(iter(os.listdir(tempdir)))
            with open(os.path.join(tempdir, temporary_file)) as f:
                assert (
                    f.read() == '"fred flintstone",35,1.28\n"barney rubble",32,2.32\n'
                )
        finally:
            con.drop_table("tempdir_sink", force=True)


def test_insert_simple_select(con, tempdir_sink_configs):
    con.create_table(
        "source",
        pd.DataFrame(
            [("fred flintstone", 35, 1.28), ("barney rubble", 32, 2.32)],
            columns=["name", "age", "gpa"],
        ),
        temp=True,
    )
    try:
        sink_schema = sch.Schema({"name": dt.string, "age": dt.int64})
        source_table = ibis.table(
            sch.Schema({"name": dt.string, "age": dt.int64, "gpa": dt.float64}),
            "source",
        )
        with tempfile.TemporaryDirectory() as tempdir:
            con.create_table(
                "tempdir_sink",
                schema=sink_schema,
                tbl_properties=tempdir_sink_configs(tempdir),
            )
            try:
                con.insert("tempdir_sink", source_table[["name", "age"]]).wait()
                temporary_file = next(iter(os.listdir(tempdir)))
                with open(os.path.join(tempdir, temporary_file)) as f:
                    assert f.read() == '"fred flintstone",35\n"barney rubble",32\n'
            finally:
                con.drop_table("tempdir_sink", force=True)
    finally:
        con.drop_table("source", force=True)


@pytest.mark.parametrize("table_name", ["new_table", None])
def test_read_csv(con, awards_players_schema, csv_source_configs, table_name):
    source_configs = csv_source_configs("awards_players")
    table = con.read_csv(
        path=source_configs["path"],
        schema=awards_players_schema,
        table_name=table_name,
    )
    try:
        if table_name is None:
            table_name = table.get_name()
        assert table_name in con.list_tables()
        assert table.schema() == awards_players_schema
    finally:
        con.drop_table(table_name)
    assert table_name not in con.list_tables()


@pytest.mark.parametrize("table_name", ["new_table", None])
def test_read_parquet(con, data_dir, tmp_path, table_name, functional_alltypes_schema):
    fname = Path("functional_alltypes.parquet")
    fname = Path(data_dir) / "parquet" / fname.name
    table = con.read_parquet(
        path=tmp_path / fname.name,
        schema=functional_alltypes_schema,
        table_name=table_name,
    )

    try:
        if table_name is None:
            table_name = table.get_name()
        assert table_name in con.list_tables()
        assert table.schema() == functional_alltypes_schema
    finally:
        con.drop_table(table_name)
    assert table_name not in con.list_tables()


@pytest.mark.parametrize("table_name", ["new_table", None])
def test_read_json(con, data_dir, tmp_path, table_name, functional_alltypes_schema):
    pq = pytest.importorskip("pyarrow.parquet")

    pq_table = pq.read_table(
        data_dir.joinpath("parquet", "functional_alltypes.parquet")
    )
    df = pq_table.to_pandas()

    path = tmp_path / "functional_alltypes.json"
    df.to_json(path, orient="records", lines=True, date_format="iso")
    table = con.read_json(
        path=path, schema=functional_alltypes_schema, table_name=table_name
    )

    try:
        if table_name is None:
            table_name = table.get_name()
        assert table_name in con.list_tables()
        assert table.schema() == functional_alltypes_schema
        assert table.count().execute() == len(pq_table)
    finally:
        con.drop_table(table_name)
    assert table_name not in con.list_tables()


@pytest.mark.parametrize(
    "table_name", ["astronauts", "awards_players", "diamonds", "functional_alltypes"]
)
def test_to_csv(con, tmp_path, table_name):
    table = con.table(table_name)
    out_path = tmp_path / "out.csv"
    con.to_csv(table, out_path)

    source_df = table.to_pandas()
    out_df = pd.read_csv(out_path)

    assert source_df.shape == out_df.shape


@pytest.mark.parametrize(
    "table_name", ["astronauts", "awards_players", "diamonds", "functional_alltypes"]
)
def test_to_parquet(con, tmp_path, table_name):
    table = con.table(table_name)
    out_path = tmp_path / "out.parquet"
    con.to_parquet(table, out_path)

    source_df = table.to_pandas()
    out_df = pd.read_parquet(out_path)

    tm.assert_frame_equal(source_df, out_df)
